import time
from uuid import UUID

import nh3
import shortuuid
from sqlalchemy.sql import text

from fittrackee import db


def clean(sql: str, days: int) -> int:
    limit = int(time.time()) - (days * 86400)
    result = db.session.execute(text(sql), {"limit": limit})
    db.session.commit()
    return result.rowcount  # type: ignore


def encode_uuid(uuid_value: UUID) -> str:
    """
    Return short id string from a UUID
    """
    return shortuuid.encode(uuid_value)


def decode_short_id(short_id: str) -> UUID:
    """
    Return UUID from a short id string
    """
    return shortuuid.decode(short_id)


def clean_input(input_text: str) -> str:
    # HTML sanitization
    return nh3.clean(
        input_text,
        tags={"a", "br", "p", "span"},
        attributes={"a": {"href", "target"}},
    )



from json import dumps
from typing import Dict, List, Optional, Union

from flask import Request, Response, current_app
from flask_sqlalchemy import SQLAlchemy

from fittrackee import appLog
from fittrackee.files import display_readable_file_size


def get_empty_data_for_datatype(data_type: str) -> Union[str, List]:
    return "" if data_type in ["gpx", "chart_data"] else []


class HttpResponse(Response):
    def __init__(
        self,
        response: Optional[Union[str, Dict]] = None,
        status_code: Optional[int] = None,
        content_type: Optional[str] = None,
    ) -> None:
        if isinstance(response, dict):
            response = dumps(response)
            content_type = (
                "application/json" if content_type is None else content_type
            )
        super().__init__(
            response=response,
            status=status_code,
            content_type=content_type,
        )


class GenericErrorResponse(HttpResponse):
    def __init__(
        self,
        status_code: int,
        message: Union[str, List],
        status: Optional[str] = None,
    ) -> None:
        response = {
            "status": "error" if status is None else status,
            "message": message,
        }
        super().__init__(
            response=response,
            status_code=status_code,
        )


class InvalidPayloadErrorResponse(GenericErrorResponse):
    def __init__(
        self,
        message: Optional[Union[str, List]] = None,
        status: Optional[str] = None,
    ) -> None:
        message = "invalid payload" if message is None else message
        super().__init__(status_code=400, message=message, status=status)


class ExceedingValueErrorResponse(InvalidPayloadErrorResponse):
    def __init__(self) -> None:
        super().__init__(
            message=(
                "one or more values, entered or calculated, exceed the limits"
            )
        )


class DataInvalidPayloadErrorResponse(HttpResponse):
    def __init__(self, data_type: str, status: Optional[str] = None) -> None:
        response = {
            "status": "error" if status is None else status,
            "data": {data_type: get_empty_data_for_datatype(data_type)},
        }
        super().__init__(response=response, status_code=400)


class InvalidConfigValueErrorResponse(HttpResponse):
    def __init__(
        self,
        key: str,
        max_value: int,
    ) -> None:
        response = {
            "status": "config_value_exceeding_limit",
            "key": key,
            "max_value": max_value,
            "message": f"'{key}' must be less than {max_value}",
        }
        super().__init__(response=response, status_code=400)


class EquipmentInvalidPayloadErrorResponse(HttpResponse):
    def __init__(
        self,
        equipment_id: str,
        message: str,
        status: Optional[str] = None,
    ) -> None:
        response = {
            "status": "error" if status is None else status,
            "equipment_id": equipment_id,
            "message": message,
        }
        super().__init__(response=response, status_code=400)


class UnauthorizedErrorResponse(GenericErrorResponse):
    def __init__(self, message: Optional[str] = None) -> None:
        message = (
            "invalid token, please request a new token"
            if message is None
            else message
        )
        super().__init__(status_code=401, message=message)


class ForbiddenErrorResponse(GenericErrorResponse):
    def __init__(self, message: Optional[str] = None) -> None:
        message = "you do not have permissions" if message is None else message
        super().__init__(status_code=403, message=message)


class NotFoundErrorResponse(GenericErrorResponse):
    def __init__(self, message: str) -> None:
        super().__init__(status_code=404, message=message, status="not found")


class UserNotFoundErrorResponse(NotFoundErrorResponse):
    def __init__(self) -> None:
        super().__init__(message="user does not exist")


class DataNotFoundErrorResponse(HttpResponse):
    def __init__(self, data_type: str, message: Optional[str] = None) -> None:
        response = {
            "status": "not found",
            "data": {data_type: get_empty_data_for_datatype(data_type)},
        }
        if message:
            response["message"] = message
        super().__init__(response=response, status_code=404)


class PayloadTooLargeErrorResponse(GenericErrorResponse):
    def __init__(
        self, file_type: str, file_size: Optional[int], max_size: Optional[int]
    ) -> None:
        readable_file_size = (
            f"({display_readable_file_size(file_size)}) " if file_size else ""
        )
        readable_max_size = (
            display_readable_file_size(max_size) if max_size else "limit"
        )
        message = (
            f"Error during {file_type} upload, file size {readable_file_size}"
            f"exceeds {readable_max_size}."
        )
        super().__init__(status_code=413, message=message, status="fail")


class InternalServerErrorResponse(GenericErrorResponse):
    def __init__(
        self, message: Optional[str] = None, status: Optional[str] = None
    ) -> None:
        message = (
            "error, please try again or contact the administrator"
            if message is None
            else message
        )
        super().__init__(status_code=500, message=message, status=status)


def handle_error_and_return_response(
    error: Exception,
    message: Optional[str] = None,
    status: Optional[str] = None,
    db: Optional[SQLAlchemy] = None,
) -> HttpResponse:
    if db is not None:
        db.session.rollback()
    appLog.error(error)
    return InternalServerErrorResponse(message=message, status=status)


def get_error_response_if_file_is_invalid(
    file_type: str, req: Request
) -> Optional[HttpResponse]:
    if "file" not in req.files:
        return InvalidPayloadErrorResponse("no file part", "fail")

    file = req.files["file"]
    if not file.filename or file.filename == "":
        return InvalidPayloadErrorResponse("no selected file", "fail")

    allowed_extensions = (
        "WORKOUT_ALLOWED_EXTENSIONS"
        if file_type == "workout"
        else "PICTURE_ALLOWED_EXTENSIONS"
    )

    file_extension = (
        file.filename.rsplit(".", 1)[1].lower()
        if "." in file.filename
        else None
    )
    max_file_size = current_app.config["max_single_file_size"]

    if not (
        file_extension
        and file_extension in current_app.config[allowed_extensions]
    ):
        return InvalidPayloadErrorResponse(
            "file extension not allowed", "fail"
        )

    if (
        file_extension != "zip"
        and req.content_length is not None
        and req.content_length > max_file_size
    ):
        return PayloadTooLargeErrorResponse(
            file_type=file_type,
            file_size=req.content_length,
            max_size=max_file_size,
        )

    return None



"""add privacy policy

Revision ID: 374a670efe23
Revises: 0f375c44e659
Create Date: 2023-02-25 11:08:08.977217

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '374a670efe23'
down_revision = '0f375c44e659'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('app_config', schema=None) as batch_op:
        batch_op.add_column(sa.Column('privacy_policy_date', sa.DateTime(), nullable=True))
        batch_op.add_column(sa.Column('privacy_policy', sa.Text(), nullable=True))
        batch_op.add_column(sa.Column('about', sa.Text(), nullable=True))

    with op.batch_alter_table('users', schema=None) as batch_op:
        batch_op.add_column(sa.Column('accepted_policy_date', sa.DateTime(), nullable=True))
        batch_op.alter_column('date_format',
               existing_type=sa.VARCHAR(length=50),
               nullable=True)

    op.create_table('users_data_export',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=True),
    sa.Column('created_at', sa.DateTime(), nullable=False),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('completed', sa.Boolean(), nullable=False),
    sa.Column('file_name', sa.String(length=100), nullable=True),
    sa.Column('file_size', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['user_id'], ['users.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id')
    )
    with op.batch_alter_table('users_data_export', schema=None) as batch_op:
        batch_op.create_index(batch_op.f('ix_users_data_export_user_id'), ['user_id'], unique=True)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('users_data_export', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_users_data_export_user_id'))

    op.drop_table('users_data_export')

    with op.batch_alter_table('users', schema=None) as batch_op:
        batch_op.alter_column('date_format',
               existing_type=sa.VARCHAR(length=50),
               nullable=False)
        batch_op.drop_column('accepted_policy_date')

    with op.batch_alter_table('app_config', schema=None) as batch_op:
        batch_op.drop_column('about')
        batch_op.drop_column('privacy_policy')
        batch_op.drop_column('privacy_policy_date')

    # ### end Alembic commands ###



from fittrackee.exceptions import GenericException


class InvalidDurationException(Exception):
    def __init__(self) -> None:
        super().__init__("invalid duration")


class InvalidGPXException(GenericException):
    pass


class WorkoutExceedingValueException(GenericException):
    def __init__(self, detail: str) -> None:
        super().__init__(
            "invalid",
            "one or more values, entered or calculated, exceed the limits",
        )
        self.detail = detail


class WorkoutException(GenericException):
    pass


class WorkoutFileException(GenericException):
    pass


class WorkoutGPXException(GenericException):
    pass


class WorkoutForbiddenException(GenericException):
    def __init__(self) -> None:
        super().__init__("error", "you do not have permissions")



from authlib.integrations.sqla_oauth2 import (
    create_bearer_token_validator,
    create_revocation_endpoint,
)
from authlib.oauth2.rfc7636 import CodeChallenge
from flask import Flask

from fittrackee import db

from .grants import AuthorizationCodeGrant, OAuth2Token, RefreshTokenGrant
from .server import authorization_server, require_auth


def config_oauth(app: Flask) -> None:
    authorization_server.init_app(app)

    # supported grants
    authorization_server.register_grant(
        AuthorizationCodeGrant, [CodeChallenge(required=True)]
    )
    authorization_server.register_grant(RefreshTokenGrant)

    # support revocation
    revocation_cls = create_revocation_endpoint(db.session, OAuth2Token)
    revocation_cls.CLIENT_AUTH_METHODS = ["client_secret_post"]
    authorization_server.register_endpoint(revocation_cls)

    # protect resource
    bearer_cls = create_bearer_token_validator(db.session, OAuth2Token)
    require_auth.register_token_validator(bearer_cls())



"""add workout description

Revision ID: 6988082918f8
Revises: c9c6a5a4dd6d
Create Date: 2024-09-07 20:02:56.341637

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '6988082918f8'
down_revision = 'c9c6a5a4dd6d'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('workouts', schema=None) as batch_op:
        batch_op.add_column(sa.Column('description', sa.String(length=10000), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('workouts', schema=None) as batch_op:
        batch_op.drop_column('description')

    # ### end Alembic commands ###



from enum import Enum
from typing import TYPE_CHECKING, Optional, Union

if TYPE_CHECKING:
    from fittrackee.comments.models import Comment
    from fittrackee.equipments.models import Equipment
    from fittrackee.users.models import User
    from fittrackee.workouts.models import Workout


class VisibilityLevel(str, Enum):  # to make enum serializable
    PUBLIC = "public"
    FOLLOWERS = "followers_only"  # only followers
    PRIVATE = "private"  # in case of comments, for mentioned users only


def get_calculated_visibility(
    *, visibility: VisibilityLevel, parent_visibility: VisibilityLevel
) -> VisibilityLevel:
    # - workout visibility overrides analysis visibility, when stricter,
    # - analysis visibility overrides map visibility, when stricter.
    if parent_visibility == VisibilityLevel.PRIVATE or (
        parent_visibility == VisibilityLevel.FOLLOWERS
        and visibility == VisibilityLevel.PUBLIC
    ):
        return parent_visibility
    return visibility


def can_view(
    target_object: Union["Workout", "Comment", "Equipment"],
    visibility: str,
    user: Optional["User"] = None,
    for_report: bool = False,
) -> bool:
    from fittrackee.comments.models import Comment

    owner = target_object.user
    if user and (
        user.id == owner.id or (user.has_moderator_rights and for_report)
    ):
        return True

    if (
        target_object.__class__.__name__ == "Workout"
        and target_object.suspended_at  # type: ignore
    ):
        if not user:
            return False
        user_comments_count = Comment.query.filter_by(
            workout_id=target_object.id, user_id=user.id
        ).count()
        if user_comments_count == 0:
            return False

    if target_object.__getattribute__(
        visibility
    ) == VisibilityLevel.PUBLIC and (
        not user or not user.is_blocked_by(owner)
    ):
        return True

    if not user:
        return False

    if (
        isinstance(target_object, Comment)
        and user in target_object.mentioned_users.all()
    ):
        return True

    if (
        target_object.__getattribute__(visibility) == VisibilityLevel.FOLLOWERS
        and user in owner.followers.all()
    ):
        return True

    # visibility level is private
    return False


def can_view_heart_rate(
    target_user: "User",
    user: Optional["User"] = None,
) -> bool:
    if user and (user.id == target_user.id):
        return True

    if target_user.hr_visibility == VisibilityLevel.PUBLIC and (
        not user or not user.is_blocked_by(target_user)
    ):
        return True

    if not user:
        return False

    if (
        target_user.hr_visibility == VisibilityLevel.FOLLOWERS
        and user in target_user.followers.all()
    ):
        return True

    # visibility level is private
    return False



from typing import IO, Optional

import fitdecode
import gpxpy.gpx

from ...exceptions import WorkoutFileException
from .constants import GARMIN_DEVICES, NSMAP
from .workout_gpx_creation_service import WorkoutGpxCreationService


class WorkoutFitCreationService(WorkoutGpxCreationService):
    @staticmethod
    def _get_coordinate(value: int) -> float:
        """
        converts coordinates from semicircles
        """
        return round(value * (180.0 / 2**31), 5)

    @classmethod
    def parse_file(cls, workout_file: IO[bytes]) -> "gpxpy.gpx.GPX":
        """
        For now only Activity Files are supported.
        see:
        https://developer.garmin.com/fit/file-types/activity/

        Activity File contains Laps (intervals in session) and Records.

        For now, only records are parsed and gpx file generated from fit file
        contains only one track. A new segment is created on after 'stop_all'
        event.

        TODO:
        - handle multiple sports activities (see Session)
        """
        try:
            fit_file = fitdecode.FitReader(workout_file)
        except Exception as e:
            raise WorkoutFileException(
                "error", "error when parsing fit file"
            ) from e

        gpx_track = gpxpy.gpx.GPXTrack()
        gpx_segment = gpxpy.gpx.GPXTrackSegment()
        has_stop = False
        creator: Optional[str] = None
        try:
            for frame in fit_file:
                if frame.frame_type != fitdecode.FIT_FRAME_DATA:
                    continue
                if frame.name == "file_id":
                    if not frame.has_field("manufacturer"):
                        continue
                    creator = frame.get_value("manufacturer")
                    if frame.has_field("product") and frame.get_value(
                        "product"
                    ):
                        product = frame.get_value("product")
                        if product in GARMIN_DEVICES.keys():
                            product = GARMIN_DEVICES[product]
                        creator = f"{creator} {product}"
                    continue
                # create a new segment after 'stop_all' event
                if (
                    frame.name == "event"
                    and frame.get_value("event") == "timer"
                    and frame.get_value("event_type") == "stop_all"
                ):
                    has_stop = True
                    gpx_track.segments.append(gpx_segment)
                    gpx_segment = gpxpy.gpx.GPXTrackSegment()
                    continue

                if frame.name != "record":
                    continue

                longitude = (
                    frame.get_value("position_long")
                    if frame.has_field("position_long")
                    else None
                )
                latitude = (
                    frame.get_value("position_lat")
                    if frame.has_field("position_lat")
                    else None
                )
                time = (
                    frame.get_value("timestamp")
                    if frame.has_field("timestamp")
                    else None
                )
                if not longitude or not latitude or not time:
                    continue

                elevation = (
                    frame.get_value("enhanced_altitude")
                    if frame.has_field("enhanced_altitude")
                    else None
                )
                heart_rate = (
                    frame.get_value("heart_rate")
                    if frame.has_field("heart_rate")
                    else None
                )
                cadence = (
                    frame.get_value("cadence")
                    if frame.has_field("cadence")
                    else None
                )

                point = gpxpy.gpx.GPXTrackPoint(
                    longitude=cls._get_coordinate(longitude),
                    latitude=cls._get_coordinate(latitude),
                    elevation=float(elevation) if elevation else None,
                    time=time,
                )

                if heart_rate is not None or cadence is not None:
                    point.extensions.append(
                        cls._get_extensions(heart_rate, cadence)
                    )
                gpx_segment.points.append(point)

        except fitdecode.exceptions.FitHeaderError as e:
            raise WorkoutFileException(
                "error", "error when parsing fit file"
            ) from e

        if not has_stop:
            gpx_track.segments.append(gpx_segment)

        gpx = gpxpy.gpx.GPX()
        gpx.creator = creator
        gpx.nsmap = NSMAP
        gpx.tracks.append(gpx_track)
        return gpx



import os
import re
import secrets
from datetime import datetime, timedelta, timezone
from typing import Dict, Tuple, Union

import jwt
from flask import (
    Blueprint,
    Response,
    current_app,
    request,
    send_from_directory,
)
from sqlalchemy import exc, func
from sqlalchemy.dialects.postgresql import insert
from werkzeug.exceptions import RequestEntityTooLarge
from werkzeug.utils import secure_filename

from fittrackee import appLog, db
from fittrackee.dates import get_datetime_in_utc, get_readable_duration
from fittrackee.emails.tasks import send_email
from fittrackee.equipments.exceptions import (
    InvalidEquipmentException,
    InvalidEquipmentsException,
)
from fittrackee.equipments.utils import handle_equipments
from fittrackee.files import get_absolute_file_path
from fittrackee.oauth2.server import require_auth
from fittrackee.reports.models import ReportAction, ReportActionAppeal
from fittrackee.responses import (
    DataNotFoundErrorResponse,
    EquipmentInvalidPayloadErrorResponse,
    ForbiddenErrorResponse,
    HttpResponse,
    InvalidPayloadErrorResponse,
    NotFoundErrorResponse,
    PayloadTooLargeErrorResponse,
    UnauthorizedErrorResponse,
    get_error_response_if_file_is_invalid,
    handle_error_and_return_response,
)
from fittrackee.users.users_service import UserManagerService
from fittrackee.utils import (
    decode_short_id,
)
from fittrackee.visibility_levels import (
    VisibilityLevel,
    get_calculated_visibility,
)
from fittrackee.workouts.models import Sport

from .exceptions import UserControlsException, UserCreationException
from .models import (
    BlacklistedToken,
    BlockedUser,
    Notification,
    User,
    UserSportPreference,
    UserSportPreferenceEquipment,
    UserTask,
)
from .roles import UserRole
from .tasks import export_data
from .timezones import TIMEZONES, get_timezone
from .utils.controls import check_password, is_valid_email
from .utils.language import get_language
from .utils.tokens import decode_user_token

auth_blueprint = Blueprint("auth", __name__)

HEX_COLOR_REGEX = regex = "^#([A-Fa-f0-9]{6}|[A-Fa-f0-9]{3})$"
NOT_FOUND_MESSAGE = "the requested URL was not found on the server"
BLOCKED_USERS_PER_PAGE = 5


def send_account_confirmation_email(user: User) -> None:
    if current_app.config["CAN_SEND_EMAILS"]:
        fittrackee_url = current_app.config["UI_URL"]
        email_data = {
            "username": user.username,
            "fittrackee_url": fittrackee_url,
            "operating_system": request.user_agent.platform,  # type: ignore
            "browser_name": request.user_agent.browser,  # type: ignore
            "account_confirmation_url": (
                f"{fittrackee_url}/account-confirmation"
                f"?token={user.confirmation_token}"
            ),
        }
        user_data = {
            "language": get_language(user.language),
            "email": user.email,
        }
        send_email.send(user_data, email_data, template="account_confirmation")


@auth_blueprint.route("/auth/register", methods=["POST"])
def register_user() -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Register a user and send confirmation email.

    The newly created account is inactive. The user must confirm his email
    to activate it.

    **Example request**:

    .. sourcecode:: http

      POST /api/auth/register HTTP/1.1
      Content-Type: application/json

    **Example responses**:

    - success:

    .. sourcecode:: http

      HTTP/1.1 200 SUCCESS
      Content-Type: application/json

      {
        "status": "success"
      }

    - error on registration:

    .. sourcecode:: http

      HTTP/1.1 400 BAD REQUEST
      Content-Type: application/json

      {
        "message": "Errors: email: valid email must be provided\\n",
        "status": "error"
      }

    :<json string username: username (3 to 30 characters required)
    :<json string email: user email
    :<json string password: password (8 characters required)
    :<json string lang: user language preferences (if not provided or invalid,
                        fallback to 'en' (english))
    :<json boolean accepted_policy: ``true`` if user accepted privacy policy
    :<json string timezone: user timezone (if not provided or invalid,
                        fallback to 'Europe/Paris')

    :statuscode 200: ``success``
    :statuscode 400:
        - ``invalid payload``
        - ``sorry, that username is already taken``
        - ``sorry, you must agree privacy policy to register``
        - ``username: 3 to 30 characters required``
        - ``username: only alphanumeric characters and the underscore
          character "_" allowed``
        - ``email: valid email must be provided``
        - ``password: 8 characters required``
    :statuscode 403: ``error, registration is disabled``
    :statuscode 500: ``error, please try again or contact the administrator``
    """
    if not current_app.config.get("is_registration_enabled"):
        return ForbiddenErrorResponse("error, registration is disabled")

    post_data = request.get_json()
    if (
        not post_data
        or post_data.get("username") is None
        or post_data.get("email") is None
        or post_data.get("password") is None
        or post_data.get("accepted_policy") is None
    ):
        return InvalidPayloadErrorResponse()

    accepted_policy = post_data.get("accepted_policy") is True
    if not accepted_policy:
        return InvalidPayloadErrorResponse(
            "sorry, you must agree privacy policy to register"
        )

    username = post_data.get("username")
    email = post_data.get("email")
    password = post_data.get("password")
    language = get_language(post_data.get("language"))
    tz = get_timezone(post_data.get("timezone"))

    try:
        user_manager_service = UserManagerService(username=username)
        new_user, _ = user_manager_service.create_user(email, password)
        # if a user exists with same email address (returned new_user is None),
        # no error is returned since a user has to confirm his email to
        # activate his account
        if new_user:
            new_user.language = language
            new_user.timezone = tz
            new_user.accepted_policy_date = datetime.now(timezone.utc)
            for admin in User.query.filter(
                User.role == UserRole.ADMIN.value,
                User.is_active == True,  # noqa
            ).all():
                if not admin.is_notification_enabled("account_creation"):
                    continue
                notification = Notification(
                    from_user_id=new_user.id,
                    to_user_id=admin.id,
                    created_at=new_user.created_at,
                    event_type="account_creation",
                    event_object_id=new_user.id,
                )
                db.session.add(notification)
            db.session.commit()
            send_account_confirmation_email(new_user)

        return {"status": "success"}, 200
    # handler errors
    except (UserControlsException, UserCreationException) as e:
        return InvalidPayloadErrorResponse(str(e))
    except (
        exc.IntegrityError,
        exc.OperationalError,
        TypeError,
        ValueError,
    ) as e:
        return handle_error_and_return_response(e, db=db)


@auth_blueprint.route("/auth/login", methods=["POST"])
def login_user() -> Union[Dict, HttpResponse]:
    """
    User login.

    Only user with an active account can log in.

    **Example request**:

    .. sourcecode:: http

      POST /api/auth/login HTTP/1.1
      Content-Type: application/json

    **Example responses**:

    - successful login:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "auth_token": "JSON Web Token",
        "message": "successfully logged in",
        "status": "success"
      }

    - error on login

    .. sourcecode:: http

      HTTP/1.1 401 UNAUTHORIZED
      Content-Type: application/json

      {
        "message": "invalid credentials",
        "status": "error"
      }

    :<json string email: user email
    :<json string password: password

    :statuscode 200: ``successfully logged in``
    :statuscode 400: ``invalid payload``
    :statuscode 401: ``invalid credentials``
    :statuscode 500: ``error, please try again or contact the administrator``

    """
    # get post data
    post_data = request.get_json()
    if not post_data:
        return InvalidPayloadErrorResponse()
    email = post_data.get("email", "")
    password = post_data.get("password")
    try:
        user = User.query.filter(
            func.lower(User.email) == func.lower(email),
            User.is_active == True,  # noqa
        ).first()
        if user and user.check_password(password):
            # generate auth token
            auth_token = user.encode_auth_token(user.id)
            return {
                "status": "success",
                "message": "successfully logged in",
                "auth_token": auth_token,
            }
        return UnauthorizedErrorResponse("invalid credentials")
    # handler errors
    except (exc.IntegrityError, exc.OperationalError, ValueError) as e:
        return handle_error_and_return_response(e, db=db)


@auth_blueprint.route("/auth/profile", methods=["GET"])
@require_auth(scopes=["profile:read"], allow_suspended_user=True)
def get_authenticated_user_profile(
    auth_user: User,
) -> Union[Dict, HttpResponse]:
    """
    Get authenticated user info (profile, account, preferences).

    Suspended user can access this endpoint.

    **Scope**: ``profile:read``

    **Example request**:

    .. sourcecode:: http

      GET /api/auth/profile HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "accepted_privacy_policy": true,
          "analysis_visibility": "private",
          "bio": null,
          "birth_date": null,
          "created_at": "Sun, 14 Jul 2019 14:09:58 GMT",
          "date_format": "dd/MM/yyyy",
          "display_ascent": true,
          "email": "sam@example.com",
          "email_to_confirm": null,
          "first_name": null,
          "followers": 0,
          "following": 0,
          "hide_profile_in_users_directory": true,
          "imperial_units": false,
          "is_active": true,
          "language": "en",
          "last_name": null,
          "location": null,
          "manually_approves_followers": false,
          "map_visibility": "private",
          "nb_sports": 3,
          "nb_workouts": 6,
          "notification_preferences": {
            "comment_like": true,
            "follow": true,
            "follow_request": true,
            "follow_request_approved": true,
            "mention": true,
            "workout_comment": true,
            "workout_like": true
          }
          "picture": false,
          "records": [
            {
              "id": 9,
              "record_type": "AS",
              "sport_id": 1,
              "user": "sam",
              "value": 18,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 10,
              "record_type": "FD",
              "sport_id": 1,
              "user": "sam",
              "value": 18,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 13,
              "record_type": "HA",
              "sport_id": 1,
              "user": "Sam",
              "value": 43.97,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 11,
              "record_type": "LD",
              "sport_id": 1,
              "user": "sam",
              "value": "1:01:00",
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 12,
              "record_type": "MS",
              "sport_id": 1,
              "user": "sam",
              "value": 18,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            }
          ],
          "role": "user",
          "sports_list": [
              1,
              4,
              6
          ],
          "start_elevation_at_zero": false,
          "timezone": "Europe/Paris",
          "total_ascent": 720.35,
          "total_distance": 67.895,
          "total_duration": "6:50:27",
          "use_dark_mode": null,
          "use_raw_gpx_speed": false,
          "username": "sam",
          "weekm": false,
          "workouts_visibility": "private"
        },
        "status": "success"
      }

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    """
    return {
        "status": "success",
        "data": auth_user.serialize(current_user=auth_user, light=False),
    }


@auth_blueprint.route("/auth/profile/edit", methods=["POST"])
@require_auth(scopes=["profile:write"], allow_suspended_user=True)
def edit_user(auth_user: User) -> Union[Dict, HttpResponse]:
    """
    Edit authenticated user profile.

    Suspended user can access this endpoint.

    **Scope**: ``profile:write``

    **Example request**:

    .. sourcecode:: http

      POST /api/auth/profile/edit HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "accepted_privacy_policy": true,
          "analysis_visibility": "private",
          "bio": null,
          "birth_date": null,
          "created_at": "Sun, 14 Jul 2019 14:09:58 GMT",
          "date_format": "dd/MM/yyyy",
          "display_ascent": true,
          "email": "sam@example.com",
          "email_to_confirm": null,
          "first_name": null,
          "followers": 0,
          "following": 0,
          "hide_profile_in_users_directory": true,
          "imperial_units": false,
          "is_active": true,
          "language": "en",
          "last_name": null,
          "location": null,
          "manually_approves_followers": false,
          "map_visibility": "private",
          "nb_sports": 3,
          "nb_workouts": 6,
          "notification_preferences": {
            "comment_like": true,
            "follow": true,
            "follow_request": true,
            "follow_request_approved": true,
            "mention": true,
            "workout_comment": true,
            "workout_like": true
          }
          "picture": false,
          "records": [
            {
              "id": 9,
              "record_type": "AS",
              "sport_id": 1,
              "user": "sam",
              "value": 18,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 10,
              "record_type": "FD",
              "sport_id": 1,
              "user": "sam",
              "value": 18,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 13,
              "record_type": "HA",
              "sport_id": 1,
              "user": "Sam",
              "value": 43.97,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 11,
              "record_type": "LD",
              "sport_id": 1,
              "user": "sam",
              "value": "1:01:00",
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 12,
              "record_type": "MS",
              "sport_id": 1,
              "user": "sam",
              "value": 18,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            }
          ],
          "role": "user",
          "sports_list": [
              1,
              4,
              6
          ],
          "start_elevation_at_zero": false,
          "timezone": "Europe/Paris",
          "total_ascent": 720.35,
          "total_distance": 67.895,
          "total_duration": "6:50:27",
          "use_dark_mode": null,
          "use_raw_gpx_speed": false,
          "username": "sam"
          "weekm": true,
          "workouts_visibility": "private"
        },
        "message": "user profile updated",
        "status": "success"
      }

    :<json string first_name: user first name
    :<json string last_name: user last name
    :<json string location: user location
    :<json string bio: user biography
    :<json string birth_date: user birth date (format: ``%Y-%m-%d``)

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``user profile updated``
    :statuscode 400: ``invalid payload``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 500: ``error, please try again or contact the administrator``
    """
    # get post data
    post_data = request.get_json()
    user_mandatory_data = {
        "first_name",
        "last_name",
        "bio",
        "birth_date",
        "location",
    }
    if not post_data or not post_data.keys() >= user_mandatory_data:
        return InvalidPayloadErrorResponse()

    first_name = post_data.get("first_name")
    last_name = post_data.get("last_name")
    bio = post_data.get("bio")
    birth_date = post_data.get("birth_date")
    location = post_data.get("location")

    try:
        auth_user.first_name = first_name
        auth_user.last_name = last_name
        auth_user.bio = bio
        auth_user.location = location
        auth_user.birth_date = (
            get_datetime_in_utc(birth_date) if birth_date else None
        )
        db.session.commit()

        return {
            "status": "success",
            "message": "user profile updated",
            "data": auth_user.serialize(current_user=auth_user, light=False),
        }

    # handler errors
    except (exc.IntegrityError, exc.OperationalError, ValueError) as e:
        return handle_error_and_return_response(e, db=db)


@auth_blueprint.route("/auth/profile/edit/account", methods=["PATCH"])
@require_auth(scopes=["profile:write"], allow_suspended_user=True)
def update_user_account(auth_user: User) -> Union[Dict, HttpResponse]:
    """
    Update authenticated user email and password.

    It sends emails if sending is enabled:

    - Password change
    - Email change:

      - one to the current address to inform user
      - another one to the new address to confirm it.

    Suspended user can access this endpoint.

    **Scope**: ``profile:write``

    **Example request**:

    .. sourcecode:: http

      PATCH /api/auth/profile/edit/account HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "accepted_privacy_policy": true,
          "analysis_visibility": "private",
          "bio": null,
          "birth_date": null,
          "created_at": "Sun, 14 Jul 2019 14:09:58 GMT",
          "date_format": "dd/MM/yyyy",
          "display_ascent": true,
          "email": "sam@example.com",
          "email_to_confirm": null,
          "first_name": null,
          "hide_profile_in_users_directory": true,
          "imperial_units": false,
          "is_active": true,
          "language": "en",
          "last_name": null,
          "location": null,
          "manually_approves_followers": false,
          "map_visibility": "followers_only",
          "nb_sports": 3,
          "nb_workouts": 6,
          "notification_preferences": {
            "comment_like": true,
            "follow": true,
            "follow_request": true,
            "follow_request_approved": true,
            "mention": true,
            "workout_comment": true,
            "workout_like": true
          }
          "picture": false,
          "records": [
            {
              "id": 9,
              "record_type": "AS",
              "sport_id": 1,
              "user": "sam",
              "value": 18,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 10,
              "record_type": "FD",
              "sport_id": 1,
              "user": "sam",
              "value": 18,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 13,
              "record_type": "HA",
              "sport_id": 1,
              "user": "Sam",
              "value": 43.97,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 11,
              "record_type": "LD",
              "sport_id": 1,
              "user": "sam",
              "value": "1:01:00",
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 12,
              "record_type": "MS",
              "sport_id": 1,
              "user": "sam",
              "value": 18,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            }
          ],
          "role": "user",
          "sports_list": [
              1,
              4,
              6
          ],
          "start_elevation_at_zero": false,
          "timezone": "Europe/Paris",
          "total_ascent": 720.35,
          "total_distance": 67.895,
          "total_duration": "6:50:27",
          "use_dark_mode": null,
          "use_raw_gpx_speed": false,
          "username": "sam"
          "weekm": true,
          "workouts_visibility": "private"
        },
        "message": "user account updated",
        "status": "success"
      }

    :<json string email: user email
    :<json string password: user current password
    :<json string new_password: user new password

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``user account updated``
    :statuscode 400:
        - ``invalid payload``
        - ``email is missing``
        - ``current password is missing``
        - ``email: valid email must be provided``
        - ``password: 8 characters required``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
        - ``invalid credentials``
    :statuscode 500: ``error, please try again or contact the administrator``
    """
    data = request.get_json()
    if not data:
        return InvalidPayloadErrorResponse()
    email_to_confirm = data.get("email")
    if not email_to_confirm:
        return InvalidPayloadErrorResponse("email is missing")
    current_password = data.get("password")
    if not current_password:
        return InvalidPayloadErrorResponse("current password is missing")
    if not auth_user.check_password(current_password):
        return UnauthorizedErrorResponse("invalid credentials")

    new_password = data.get("new_password")
    error_messages = ""
    try:
        if email_to_confirm != auth_user.email:
            if is_valid_email(email_to_confirm):
                if current_app.config["CAN_SEND_EMAILS"]:
                    auth_user.email_to_confirm = email_to_confirm
                    auth_user.confirmation_token = secrets.token_urlsafe(30)
                else:
                    auth_user.email = email_to_confirm
                    auth_user.confirmation_token = None
            else:
                error_messages = "email: valid email must be provided\n"

        if new_password is not None:
            error_messages += check_password(new_password)
            if error_messages == "":
                hashed_password = auth_user.generate_password_hash(
                    new_password
                )
                auth_user.password = hashed_password

        if error_messages != "":
            return InvalidPayloadErrorResponse(error_messages)

        db.session.commit()

        if current_app.config["CAN_SEND_EMAILS"]:
            fittrackee_url = current_app.config["UI_URL"]
            user_data = {
                "language": get_language(auth_user.language),
                "email": auth_user.email,
            }
            data = {
                "username": auth_user.username,
                "fittrackee_url": fittrackee_url,
                "operating_system": request.user_agent.platform,
                "browser_name": request.user_agent.browser,
            }

            if new_password is not None:
                send_email.send(user_data, data, template="password_change")

            if (
                auth_user.email_to_confirm is not None
                and auth_user.email_to_confirm != auth_user.email
            ):
                email_data = {
                    **data,
                    **{"new_email_address": email_to_confirm},
                }
                send_email.send(
                    user_data,
                    email_data,
                    template="email_update_to_current_email",
                )

                email_data = {
                    **data,
                    **{
                        "email_confirmation_url": (
                            f"{fittrackee_url}/email-update"
                            f"?token={auth_user.confirmation_token}"
                        )
                    },
                }
                user_data = {
                    **user_data,
                    **{"email": auth_user.email_to_confirm},
                }
                send_email.send(
                    user_data, email_data, template="email_update_to_new_email"
                )

        return {
            "status": "success",
            "message": "user account updated",
            "data": auth_user.serialize(current_user=auth_user, light=False),
        }

    except (exc.IntegrityError, exc.OperationalError, ValueError) as e:
        return handle_error_and_return_response(e, db=db)


@auth_blueprint.route("/auth/profile/edit/preferences", methods=["POST"])
@require_auth(scopes=["profile:write"], allow_suspended_user=True)
def edit_user_preferences(auth_user: User) -> Union[Dict, HttpResponse]:
    """
    Edit authenticated user preferences.

    Supported date formats:

    - ``MM/dd/yyyy`` (default value)
    - ``dd/MM/yyyy``
    - ``yyyy-MM-dd``
    - ``date_string``, corresponding on client to:

      - ``MMM. do, yyyy`` for ``en`` locale
      - ``d MMM yyyy`` for ``es``, ``fr``, ``gl``, ``it`` and ``nl`` locales
      - ``do MMM yyyy`` for ``de`` and ``nb`` locales

    Suspended user can access this endpoint.

    **Scope**: ``profile:write``

    **Example request**:

    .. sourcecode:: http

      POST /api/auth/profile/edit/preferences HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "accepted_privacy_policy": true,
          "analysis_visibility": "private",
          "bio": null,
          "birth_date": null,
          "created_at": "Sun, 14 Jul 2019 14:09:58 GMT",
          "date_format": "MM/dd/yyyy",
          "display_ascent": true,
          "email": "sam@example.com",
          "email_to_confirm": null,
          "first_name": null,
          "followers": 0,
          "following": 0,
          "hide_profile_in_users_directory": true,
          "imperial_units": false,
          "is_active": true,
          "language": "en",
          "last_name": null,
          "location": null,
          "manually_approves_followers": false,
          "map_visibility": "followers_only",
          "nb_sports": 3,
          "nb_workouts": 6,
          "notification_preferences": {
            "comment_like": true,
            "follow": true,
            "follow_request": true,
            "follow_request_approved": true,
            "mention": true,
            "workout_comment": true,
            "workout_like": true
          }
          "picture": false,
          "records": [
            {
              "id": 9,
              "record_type": "AS",
              "sport_id": 1,
              "user": "sam",
              "value": 18,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 10,
              "record_type": "FD",
              "sport_id": 1,
              "user": "sam",
              "value": 18,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 13,
              "record_type": "HA",
              "sport_id": 1,
              "user": "Sam",
              "value": 43.97,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 11,
              "record_type": "LD",
              "sport_id": 1,
              "user": "sam",
              "value": "1:01:00",
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 12,
              "record_type": "MS",
              "sport_id": 1,
              "user": "sam",
              "value": 18,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            }
          ],
          "role": "user",
          "sports_list": [
              1,
              4,
              6
          ],
          "start_elevation_at_zero": true,
          "timezone": "Europe/Paris",
          "total_ascent": 720.35,
          "total_distance": 67.895,
          "total_duration": "6:50:27",
          "use_dark_mode": null,
          "use_raw_gpx_speed": true,
          "username": "sam"
          "weekm": true,
          "workouts_visibility": "public"
        },
        "message": "user preferences updated",
        "status": "success"
      }

    :<json string analysis_visibility: workout analysis visibility
                  (``public``, ``followers_only``, ``private``)
    :<json string date_format: the format used to display dates in the app
    :<json boolean display_ascent: display highest ascent records and total
    :<json boolean hide_profile_in_users_directory: if ``true``, user does not
                                                    appear in users directory
    :<json boolean imperial_units: display distance in imperial units
    :<json string language: language preferences
    :<json string map_visibility: workout map visibility
                  (``public``, ``followers_only``, ``private``)
    :<json boolean manually_approves_followers: if ``false``, follow requests
                  are automatically approved
    :<json boolean start_elevation_at_zero: do elevation plots start at zero?
    :<json string timezone: user time zone
    :<json boolean use_dark_mode: Display interface with dark mode if ``true``.
                   If ``null``, it uses browser preferences.
    :<json boolean use_raw_gpx_speed: Use unfiltered gpx to calculate speeds
    :<json boolean weekm: does week start on Monday?
    :<json string workouts_visibility: user workouts visibility
                  (``public``, ``followers_only``, ``private``)

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``user preferences updated``
    :statuscode 400:
        - ``invalid payload``
        - ``password: password and password confirmation don't match``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 500: ``error, please try again or contact the administrator``
    """
    # get post data
    post_data = request.get_json()
    user_mandatory_data = {
        "analysis_visibility",
        "date_format",
        "display_ascent",
        "hide_profile_in_users_directory",
        "hr_visibility",
        "imperial_units",
        "language",
        "manually_approves_followers",
        "map_visibility",
        "start_elevation_at_zero",
        "timezone",
        "use_dark_mode",
        "use_raw_gpx_speed",
        "weekm",
        "workouts_visibility",
    }
    if not post_data or not post_data.keys() >= user_mandatory_data:
        return InvalidPayloadErrorResponse()

    date_format = post_data.get("date_format")
    display_ascent = post_data.get("display_ascent")
    imperial_units = post_data.get("imperial_units")
    language = get_language(post_data.get("language"))
    start_elevation_at_zero = post_data.get("start_elevation_at_zero")
    use_raw_gpx_speed = post_data.get("use_raw_gpx_speed")
    use_dark_mode = post_data.get("use_dark_mode")
    timezone = post_data.get("timezone")
    weekm = post_data.get("weekm")
    map_visibility = post_data.get("map_visibility")
    analysis_visibility = post_data.get("analysis_visibility")
    workouts_visibility = post_data.get("workouts_visibility")
    manually_approves_followers = post_data.get("manually_approves_followers")
    hide_profile_in_users_directory = post_data.get(
        "hide_profile_in_users_directory"
    )
    hr_visibility = post_data.get("hr_visibility")

    try:
        auth_user.date_format = date_format
        auth_user.display_ascent = display_ascent
        auth_user.imperial_units = imperial_units
        auth_user.language = language
        auth_user.start_elevation_at_zero = start_elevation_at_zero
        auth_user.timezone = timezone
        auth_user.use_dark_mode = use_dark_mode
        auth_user.use_raw_gpx_speed = use_raw_gpx_speed
        auth_user.weekm = weekm
        auth_user.workouts_visibility = VisibilityLevel(workouts_visibility)
        auth_user.analysis_visibility = get_calculated_visibility(
            visibility=VisibilityLevel(analysis_visibility),
            parent_visibility=auth_user.workouts_visibility,
        )
        auth_user.map_visibility = get_calculated_visibility(
            visibility=VisibilityLevel(map_visibility),
            parent_visibility=auth_user.analysis_visibility,
        )
        auth_user.manually_approves_followers = manually_approves_followers
        auth_user.hide_profile_in_users_directory = (
            hide_profile_in_users_directory
        )
        auth_user.hr_visibility = VisibilityLevel(hr_visibility)
        db.session.commit()

        return {
            "status": "success",
            "message": "user preferences updated",
            "data": auth_user.serialize(current_user=auth_user, light=False),
        }

    # handler errors
    except (exc.IntegrityError, exc.OperationalError, ValueError) as e:
        return handle_error_and_return_response(e, db=db)


@auth_blueprint.route("/auth/profile/edit/sports", methods=["POST"])
@require_auth(scopes=["profile:write"])
def edit_user_sport_preferences(
    auth_user: User,
) -> Union[Dict, HttpResponse]:
    """
    Edit authenticated user sport preferences.

    **Scope**: ``profile:write``

    **Example request**:

    .. sourcecode:: http

      POST /api/auth/profile/edit/sports HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "color": "#000000",
          "default_equipment_ids": [],
          "is_active": true,
          "sport_id": 1,
          "stopped_speed_threshold": 1,
          "user_id": 1
        },
        "message": "user sport preferences updated",
        "status": "success"
      }

    :<json int sport_id: id of the sport for which preferences are
           created/modified
    :<json string color: valid hexadecimal color
    :<json boolean is_active: is sport available when adding a workout
    :<json float stopped_speed_threshold: stopped speed threshold used by gpxpy
    :<json array of strings default_equipment_ids: the default equipment id
           to use for this sport.
           **Note**: for now only one equipment can be associated.

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``user sport preferences updated``
    :statuscode 400:
        - ``invalid payload``
        - ``invalid hexadecimal color``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
        - ``equipment_ids must be an array of strings``
        - ``only one equipment can be added``
        - ``equipment with id <equipment_id> does not exist``
        - ``invalid equipment id <equipment_id> for sport``
        - ``equipment with id <equipment_id> is inactive``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``
    :statuscode 404: ``sport does not exist``
    :statuscode 500: ``error, please try again or contact the administrator``
    """
    post_data = request.get_json()
    if (
        not post_data
        or "sport_id" not in post_data
        or len(post_data.keys()) == 1
    ):
        return InvalidPayloadErrorResponse()

    sport_id = post_data.get("sport_id")
    sport = Sport.query.filter_by(id=sport_id).first()
    if not sport:
        return NotFoundErrorResponse("sport does not exist")

    color = post_data.get("color")
    is_active = post_data.get("is_active")
    stopped_speed_threshold = post_data.get("stopped_speed_threshold")
    default_equipment_ids = post_data.get("default_equipment_ids")

    try:
        user_sport = UserSportPreference.query.filter_by(
            user_id=auth_user.id,
            sport_id=sport_id,
        ).first()
        if not user_sport:
            user_sport = UserSportPreference(
                user_id=auth_user.id,
                sport_id=sport_id,
                stopped_speed_threshold=sport.stopped_speed_threshold,
            )
            db.session.add(user_sport)
            db.session.flush()
        if color:
            if re.match(HEX_COLOR_REGEX, color) is None:
                return InvalidPayloadErrorResponse("invalid hexadecimal color")
            user_sport.color = color
        if is_active is not None:
            user_sport.is_active = is_active
        if stopped_speed_threshold:
            user_sport.stopped_speed_threshold = stopped_speed_threshold

        if default_equipment_ids is not None:
            existing_default_equipments = user_sport.default_equipments.all()
            default_equipments = handle_equipments(
                default_equipment_ids,
                auth_user,
                sport_id,
                existing_default_equipments,
            )
            if default_equipments:
                db.session.execute(
                    insert(UserSportPreferenceEquipment)
                    .values(
                        [
                            {
                                "equipment_id": equipment.id,
                                "sport_id": user_sport.sport_id,
                                "user_id": auth_user.id,
                            }
                            for equipment in default_equipments
                        ]
                    )
                    .on_conflict_do_nothing()
                )

                equipments_to_remove = set(existing_default_equipments) - set(
                    default_equipments
                )
                db.session.query(UserSportPreferenceEquipment).filter(
                    UserSportPreferenceEquipment.c.user_id == auth_user.id,
                    (
                        UserSportPreferenceEquipment.c.sport_id
                        == user_sport.sport_id
                    ),
                    UserSportPreferenceEquipment.c.equipment_id.in_(
                        [e.id for e in equipments_to_remove]
                    ),
                ).delete()

            elif existing_default_equipments:
                db.session.query(UserSportPreferenceEquipment).filter(
                    UserSportPreferenceEquipment.c.user_id == auth_user.id,
                    UserSportPreferenceEquipment.c.sport_id
                    == user_sport.sport_id,
                ).delete()
        db.session.commit()

        return {
            "status": "success",
            "message": "user sport preferences updated",
            "data": user_sport.serialize(),
        }

    # handler errors
    except InvalidEquipmentsException as e:
        return InvalidPayloadErrorResponse(str(e))
    except InvalidEquipmentException as e:
        return EquipmentInvalidPayloadErrorResponse(
            equipment_id=e.equipment_id, message=e.message, status=e.status
        )
    except (exc.IntegrityError, exc.OperationalError, ValueError) as e:
        return handle_error_and_return_response(e, db=db)


@auth_blueprint.route("/auth/profile/edit/notifications", methods=["POST"])
@require_auth(scopes=["profile:write"])
def edit_user_notifications_preferences(
    auth_user: User,
) -> Union[Dict, HttpResponse]:
    """
    Edit authenticated user preferences for UI notifications.

    **Scope**: ``profile:write``

    **Example request**:

    .. sourcecode:: http

      POST /api/auth/profile/edit/preferences HTTP/1.1
      Content-Type: application/json

    **Example responses**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "data": {
          "accepted_privacy_policy": true,
          "analysis_visibility": "private",
          "bio": null,
          "birth_date": null,
          "created_at": "Sun, 14 Jul 2019 14:09:58 GMT",
          "date_format": "dd/MM/yyyy",
          "display_ascent": true,
          "email": "sam@example.com",
          "email_to_confirm": null,
          "first_name": null,
          "followers": 0,
          "following": 0,
          "hide_profile_in_users_directory": true,
          "imperial_units": false,
          "is_active": true,
          "language": "en",
          "last_name": null,
          "location": null,
          "manually_approves_followers": false,
          "map_visibility": "private",
          "nb_sports": 3,
          "nb_workouts": 6,
          "notification_preferences": {
            "comment_like": true,
            "follow": true,
            "follow_request": true,
            "follow_request_approved": true,
            "mention": false,
            "workout_comment": false,
            "workout_like": false
          }
          "picture": false,
          "records": [
            {
              "id": 9,
              "record_type": "AS",
              "sport_id": 1,
              "user": "sam",
              "value": 18,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 10,
              "record_type": "FD",
              "sport_id": 1,
              "user": "sam",
              "value": 18,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 13,
              "record_type": "HA",
              "sport_id": 1,
              "user": "Sam",
              "value": 43.97,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 11,
              "record_type": "LD",
              "sport_id": 1,
              "user": "sam",
              "value": "1:01:00",
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            },
            {
              "id": 12,
              "record_type": "MS",
              "sport_id": 1,
              "user": "sam",
              "value": 18,
              "workout_date": "Sun, 07 Jul 2019 08:00:00 GMT",
              "workout_id": "hvYBqYBRa7wwXpaStWR4V2"
            }
          ],
          "sports_list": [
              1,
              4,
              6
          ],
          "start_elevation_at_zero": false,
          "timezone": "Europe/Paris",
          "total_ascent": 720.35,
          "total_distance": 67.895,
          "total_duration": "6:50:27",
          "use_dark_mode": null,
          "use_raw_gpx_speed": false,
          "username": "sam",
          "weekm": false,
          "workouts_visibility": "private"
        },
        "status": "success"
      }

    :<json boolean account_creation: notification for user registration
           (only for user with administration rights)
    :<json boolean comment_like: notification for comment likes
    :<json boolean follow: notification for follow
    :<json boolean follow_request: notification for follow requests
    :<json boolean follow_request_approved: notification for follow request
           approval
    :<json boolean mention: notification for mention
    :<json boolean workout_comment: notification for comments on workout
    :<json boolean workout_like: notification for workout likes

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``user preferences updated``
    :statuscode 400:
        - ``invalid payload``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``
    :statuscode 500: ``error, please try again or contact the administrator``
    """
    preferences_data = request.get_json()
    mandatory_data = {
        "comment_like",
        "follow",
        "follow_request",
        "follow_request_approved",
        "mention",
        "workout_comment",
        "workout_like",
    }
    if not preferences_data or not preferences_data.keys() >= mandatory_data:
        return InvalidPayloadErrorResponse()
    if (
        auth_user.role >= UserRole.ADMIN.value
        and "account_creation" not in preferences_data
    ):
        return InvalidPayloadErrorResponse()

    auth_user.update_preferences(preferences_data)
    db.session.commit()

    return {
        "data": auth_user.serialize(current_user=auth_user, light=False),
        "status": "success",
    }


@auth_blueprint.route(
    "/auth/profile/reset/sports/<sport_id>", methods=["DELETE"]
)
@require_auth(scopes=["profile:write"])
def reset_user_sport_preferences(
    auth_user: User, sport_id: int
) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Reset authenticated user preferences for a given sport.

    **Scope**: ``profile:write``

    **Example request**:

    .. sourcecode:: http

      DELETE /api/auth/profile/reset/sports/1 HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 204 OK
      Content-Type: application/json

    :param string sport_id: sport id

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 204: user preferences deleted
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``
    :statuscode 404: ``sport does not exist``
    :statuscode 500: ``error, please try again or contact the administrator``
    """
    sport = Sport.query.filter_by(id=sport_id).first()
    if not sport:
        return NotFoundErrorResponse("sport does not exist")

    try:
        user_sport = UserSportPreference.query.filter_by(
            user_id=auth_user.id,
            sport_id=sport_id,
        ).first()
        if user_sport:
            db.session.delete(user_sport)
            db.session.commit()
        return {"status": "no content"}, 204

    # handler errors
    except (exc.IntegrityError, exc.OperationalError) as e:
        return handle_error_and_return_response(e, db=db)


@auth_blueprint.route("/auth/picture", methods=["POST"])
@require_auth(scopes=["profile:write"], allow_suspended_user=True)
def edit_picture(auth_user: User) -> Union[Dict, HttpResponse]:
    """
    Update authenticated user picture.

    Suspended user can access this endpoint.

    **Scope**: ``profile:write``

    **Example request**:

    .. sourcecode:: http

      POST /api/auth/picture HTTP/1.1
      Content-Type: multipart/form-data

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "message": "user picture updated",
        "status": "success"
      }

    :form file: image file (allowed extensions: .jpg, .png, .gif)

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``user picture updated``
    :statuscode 400:
        - ``invalid payload``
        - ``no file part``
        - ``no selected file``
        - ``file extension not allowed``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 413: ``error during picture update: file size exceeds 1.0MB``
    :statuscode 500: ``error during picture update``
    """
    try:
        response_object = get_error_response_if_file_is_invalid(
            "picture", request
        )
    except RequestEntityTooLarge as e:
        appLog.error(e)
        return PayloadTooLargeErrorResponse(
            file_type="picture",
            file_size=request.content_length,
            max_size=current_app.config["MAX_CONTENT_LENGTH"],
        )
    if response_object:
        return response_object

    file = request.files["file"]
    filename = secure_filename(file.filename)  # type: ignore
    dirpath = os.path.join(
        current_app.config["UPLOAD_FOLDER"], "pictures", str(auth_user.id)
    )
    if not os.path.exists(dirpath):
        os.makedirs(dirpath)
    absolute_picture_path = os.path.join(dirpath, filename)
    relative_picture_path = os.path.join(
        "pictures", str(auth_user.id), filename
    )

    try:
        if auth_user.picture is not None:
            old_picture_path = get_absolute_file_path(auth_user.picture)
            if os.path.isfile(get_absolute_file_path(old_picture_path)):
                os.remove(old_picture_path)
        file.save(absolute_picture_path)
        auth_user.picture = relative_picture_path
        db.session.commit()
        return {
            "status": "success",
            "message": "user picture updated",
        }

    except (exc.IntegrityError, ValueError) as e:
        return handle_error_and_return_response(
            e, message="error during picture update", status="fail", db=db
        )


@auth_blueprint.route("/auth/picture", methods=["DELETE"])
@require_auth(scopes=["profile:write"], allow_suspended_user=True)
def del_picture(auth_user: User) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Delete authenticated user picture.

    Suspended user can access this endpoint.

    **Scope**: ``profile:write``

    **Example request**:

    .. sourcecode:: http

      DELETE /api/auth/picture HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 204 NO CONTENT
      Content-Type: application/json

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 204: picture deleted
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 500: ``error during picture deletion``

    """
    if not auth_user.picture:
        return {"status": "no content"}, 204

    try:
        picture_path = get_absolute_file_path(auth_user.picture)
        if os.path.isfile(picture_path):
            os.remove(picture_path)
        auth_user.picture = None
        db.session.commit()
        return {"status": "no content"}, 204
    except (exc.IntegrityError, ValueError) as e:
        return handle_error_and_return_response(
            e, message="error during picture deletion", status="fail", db=db
        )


@auth_blueprint.route("/auth/password/reset-request", methods=["POST"])
def request_password_reset() -> Union[Dict, HttpResponse]:
    """
    Handle password reset request.

    If email sending is disabled, this endpoint is not available.

    **Example request**:

    .. sourcecode:: http

      POST /api/auth/password/reset-request HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "message": "password reset request processed",
        "status": "success"
      }

    :<json string email: user email

    :statuscode 200: ``password reset request processed``
    :statuscode 400: ``invalid payload``
    :statuscode 404: ``the requested URL was not found on the server``

    """
    if not current_app.config["CAN_SEND_EMAILS"]:
        return NotFoundErrorResponse(NOT_FOUND_MESSAGE)

    post_data = request.get_json()
    if not post_data or post_data.get("email") is None:
        return InvalidPayloadErrorResponse()
    email = post_data.get("email")

    user = User.query.filter(User.email == email).first()
    if user:
        password_reset_token = user.encode_password_reset_token(user.id)
        fittrackee_url = current_app.config["UI_URL"]
        user_language = get_language(user.language)
        email_data = {
            "expiration_delay": get_readable_duration(
                current_app.config["PASSWORD_TOKEN_EXPIRATION_SECONDS"],
                user_language,
            ),
            "username": user.username,
            "password_reset_url": (
                f"{fittrackee_url}/password-reset?token={password_reset_token}"
            ),
            "fittrackee_url": fittrackee_url,
            "operating_system": request.user_agent.platform,  # type: ignore
            "browser_name": request.user_agent.browser,  # type: ignore
        }
        user_data = {
            "language": user_language,
            "email": user.email,
        }
        send_email.send(
            user_data, email_data, template="password_reset_request"
        )
    return {
        "status": "success",
        "message": "password reset request processed",
    }


@auth_blueprint.route("/auth/password/update", methods=["POST"])
def update_password() -> Union[Dict, HttpResponse]:
    """
    Update user password after password reset request.

    It sends emails if sending is enabled.

    **Example request**:

    .. sourcecode:: http

      POST /api/auth/password/update HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "message": "password updated",
        "status": "success"
      }

    :<json string password: password (8 characters required)
    :<json string token: password reset token

    :statuscode 200: ``password updated``
    :statuscode 400: ``invalid payload``
    :statuscode 401: ``invalid token, please request a new token``
    :statuscode 500: ``error, please try again or contact the administrator``

    """
    post_data = request.get_json()
    if (
        not post_data
        or post_data.get("password") is None
        or post_data.get("token") is None
    ):
        return InvalidPayloadErrorResponse()
    password = post_data.get("password")
    token = post_data.get("token")

    try:
        user_id = decode_user_token(token)
    except (jwt.ExpiredSignatureError, jwt.InvalidTokenError):
        return UnauthorizedErrorResponse()

    message = check_password(password)
    if message != "":
        return InvalidPayloadErrorResponse(message)

    user = User.query.filter(User.id == user_id).first()
    if not user:
        return UnauthorizedErrorResponse()
    try:
        user.password = user.generate_password_hash(password)
        db.session.commit()

        if current_app.config["CAN_SEND_EMAILS"]:
            send_email.send(
                user_data={
                    "language": get_language(user.language),
                    "email": user.email,
                },
                email_data={
                    "username": user.username,
                    "fittrackee_url": current_app.config["UI_URL"],
                    "operating_system": request.user_agent.platform,
                    "browser_name": request.user_agent.browser,
                },
                template="password_change",
            )

        return {
            "status": "success",
            "message": "password updated",
        }
    except (exc.OperationalError, ValueError) as e:
        return handle_error_and_return_response(e, db=db)


@auth_blueprint.route("/auth/email/update", methods=["POST"])
def update_email() -> Union[Dict, HttpResponse]:
    """
    Update user email after confirmation.

    **Example request**:

    .. sourcecode:: http

      POST /api/auth/email/update HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "message": "email updated",
        "status": "success"
      }

    :<json string token: password reset token

    :statuscode 200: ``email updated``
    :statuscode 400: ``invalid payload``
    :statuscode 500: ``error, please try again or contact the administrator``

    """
    post_data = request.get_json()
    if not post_data or post_data.get("token") is None:
        return InvalidPayloadErrorResponse()
    token = post_data.get("token")

    try:
        user = User.query.filter_by(confirmation_token=token).first()

        if not user:
            return InvalidPayloadErrorResponse()

        user.email = user.email_to_confirm
        user.email_to_confirm = None
        user.confirmation_token = None

        db.session.commit()

        response = {
            "status": "success",
            "message": "email updated",
        }

        return response

    except (exc.OperationalError, ValueError) as e:
        return handle_error_and_return_response(e, db=db)


@auth_blueprint.route("/auth/account/confirm", methods=["POST"])
def confirm_account() -> Union[Dict, HttpResponse]:
    """
    Activate user account after registration.

    **Example request**:

    .. sourcecode:: http

      POST /api/auth/account/confirm HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "auth_token": "JSON Web Token",
        "message": "account confirmation successful",
        "status": "success"
      }

    :<json string token: confirmation token

    :statuscode 200: ``account confirmation successful``
    :statuscode 400: ``invalid payload``
    :statuscode 500: ``error, please try again or contact the administrator``

    """
    post_data = request.get_json()
    if not post_data or post_data.get("token") is None:
        return InvalidPayloadErrorResponse()
    token = post_data.get("token")

    try:
        user = User.query.filter_by(confirmation_token=token).first()

        if not user:
            return InvalidPayloadErrorResponse()

        user.is_active = True
        user.confirmation_token = None

        db.session.commit()

        # generate auth token
        auth_token = user.encode_auth_token(user.id)

        response = {
            "status": "success",
            "message": "account confirmation successful",
            "auth_token": auth_token,
        }
        return response

    except (exc.OperationalError, ValueError) as e:
        return handle_error_and_return_response(e, db=db)


@auth_blueprint.route("/auth/account/resend-confirmation", methods=["POST"])
def resend_account_confirmation_email() -> Union[Dict, HttpResponse]:
    """
    Resend email with instructions to confirm account.

    If email sending is disabled, this endpoint is not available.

    **Example request**:

    .. sourcecode:: http

      POST /api/auth/account/resend-confirmation HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "message": "confirmation email resent",
        "status": "success"
      }

    :<json string email: user email

    :statuscode 200: ``confirmation email resent``
    :statuscode 400: ``invalid payload``
    :statuscode 404: ``the requested URL was not found on the server``
    :statuscode 500: ``error, please try again or contact the administrator``

    """
    if not current_app.config["CAN_SEND_EMAILS"]:
        return NotFoundErrorResponse(NOT_FOUND_MESSAGE)

    post_data = request.get_json()
    if not post_data or post_data.get("email") is None:
        return InvalidPayloadErrorResponse()
    email = post_data.get("email")

    try:
        user = User.query.filter_by(email=email, is_active=False).first()
        if user:
            user.confirmation_token = secrets.token_urlsafe(30)
            db.session.commit()

            send_account_confirmation_email(user)

        response = {
            "status": "success",
            "message": "confirmation email resent",
        }
        return response
    except (exc.OperationalError, ValueError) as e:
        return handle_error_and_return_response(e, db=db)


@auth_blueprint.route("/auth/logout", methods=["POST"])
@require_auth(allow_suspended_user=True)
def logout_user(auth_user: User) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    User logout.
    If a valid token is provided, it will be blacklisted.

    Suspended user can access this endpoint.

    **Example request**:

    .. sourcecode:: http

      POST /api/auth/logout HTTP/1.1
      Content-Type: application/json

    **Example responses**:

    - successful logout:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "message": "successfully logged out",
        "status": "success"
      }

    - error on logout:

    .. sourcecode:: http

      HTTP/1.1 401 UNAUTHORIZED
      Content-Type: application/json

      {
        "message": "provide a valid auth token",
        "status": "error"
      }

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``successfully logged out``
    :statuscode 401:
      - ``provide a valid auth token``
      - ``The access token provided is expired, revoked, malformed, or invalid
        for other reasons.``
    :statuscode 500: ``error on token blacklist``

    """
    auth_token = request.headers.get("Authorization", "").split(" ")[1]
    try:
        db.session.add(BlacklistedToken(token=auth_token))
        db.session.commit()
    except Exception:
        return {
            "status": "error",
            "message": "error on token blacklist",
        }, 500

    return {
        "status": "success",
        "message": "successfully logged out",
    }, 200


@auth_blueprint.route("/auth/account/privacy-policy", methods=["POST"])
@require_auth(allow_suspended_user=True)
def accept_privacy_policy(auth_user: User) -> Union[Dict, HttpResponse]:
    """
    The authenticated user accepts the privacy policy.

    Suspended user can access this endpoint.

    **Example request**:

    .. sourcecode:: http

      POST /api/auth/account/privacy-policy HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "status": "success"
      }

    :<json boolean accepted_policy: ``true`` if user accepted privacy policy

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 400: ``invalid payload``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 500: ``error, please try again or contact the administrator``
    """
    post_data = request.get_json()
    if not post_data or not post_data.get("accepted_policy"):
        return InvalidPayloadErrorResponse()

    try:
        if post_data.get("accepted_policy") is True:
            auth_user.accepted_policy_date = datetime.now(timezone.utc)
            db.session.commit()
            return {"status": "success"}
        else:
            return InvalidPayloadErrorResponse()
    except (exc.IntegrityError, exc.OperationalError, ValueError) as e:
        return handle_error_and_return_response(e, db=db)


@auth_blueprint.route("/auth/account/export/request", methods=["POST"])
@require_auth(allow_suspended_user=True)
def request_user_data_export(auth_user: User) -> Union[Dict, HttpResponse]:
    """
    Request a data export for authenticated user.

    Suspended user can access this endpoint.

    **Example request**:

    .. sourcecode:: http

      POST /api/auth/account/export/request HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "status": "success",
        "request": {
            "created_at": "Wed, 01 Mar 2023 12:31:17 GMT",
            "status": "in_progress",
            "file_name": null,
            "file_size": null
        }
      }

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 400:
        - ``ongoing request exists``
        - ``completed request already exists``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 500: ``error, please try again or contact the administrator``
    """
    existing_export_request = UserTask.query.filter_by(
        task_type="user_data_export", user_id=auth_user.id
    ).first()
    if existing_export_request:
        if not existing_export_request.completed:
            return InvalidPayloadErrorResponse("ongoing request exists")

        export_expiration = current_app.config["DATA_EXPORT_EXPIRATION"]
        if existing_export_request.created_at > (
            datetime.now(timezone.utc) - timedelta(hours=export_expiration)
        ):
            return InvalidPayloadErrorResponse(
                "completed request already exists"
            )

    try:
        if existing_export_request:
            db.session.delete(existing_export_request)
            db.session.flush()
        export_request = UserTask(
            user_id=auth_user.id, task_type="user_data_export"
        )
        db.session.add(export_request)
        db.session.commit()

        export_data.send(task_id=export_request.id)

        return {
            "status": "success",
            "request": export_request.serialize(current_user=auth_user),
        }
    except (exc.IntegrityError, exc.OperationalError, ValueError) as e:
        return handle_error_and_return_response(e, db=db)


@auth_blueprint.route("/auth/account/export", methods=["GET"])
@require_auth(allow_suspended_user=True)
def get_user_data_export(auth_user: User) -> Union[Dict, HttpResponse]:
    """
    Get a data export info for authenticated user if a request exists.

    It returns:

    - export creation date
    - export status (``in_progress``, ``successful`` and ``errored``)
    - file name and size (in bytes) when export is successful

    Suspended user can access this endpoint.

    **Example request**:

    .. sourcecode:: http

      GET /api/auth/account/export HTTP/1.1
      Content-Type: application/json

    **Example response**:

    - if a request exists:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "status": "success",
        "request": {
            "created_at": "Wed, 01 Mar 2023 12:31:17 GMT",
            "status": "successful",
            "file_name": "archive_rgjsR3fHt295ywNQr5Yp.zip",
            "file_size": 924
        }
      }

    - if no request:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "status": "success",
        "request": null
      }

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    """
    export_request = UserTask.query.filter_by(
        user_id=auth_user.id, task_type="user_data_export"
    ).first()
    return {
        "status": "success",
        "request": export_request.serialize(current_user=auth_user)
        if export_request
        else None,
    }


@auth_blueprint.route(
    "/auth/account/export/<string:file_name>", methods=["GET"]
)
@require_auth(allow_suspended_user=True)
def download_data_export(
    auth_user: User, file_name: str
) -> Union[Response, HttpResponse]:
    """
    Download a data export archive.

    Suspended user can access this endpoint.

    **Example request**:

    .. sourcecode:: http

      GET /api/auth/account/export/download/archive_rgjsR3fHr5Yp.zip HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/x-gzip

    :param string file_name: filename

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 404: ``file not found``
    """
    export_request = UserTask.query.filter_by(
        user_id=auth_user.id, task_type="user_data_export"
    ).first()
    if (
        not export_request
        or not export_request.completed
        or export_request.file_path.split("/")[-1] != file_name
    ):
        return DataNotFoundErrorResponse(
            data_type="archive", message="file not found"
        )

    return send_from_directory(
        current_app.config["UPLOAD_FOLDER"],
        export_request.file_path,
        mimetype="application/zip",
        as_attachment=True,
    )


@auth_blueprint.route("/auth/blocked-users", methods=["GET"])
@require_auth(scopes=["profile:read"])
def get_blocked_users(auth_user: User) -> Union[Dict, HttpResponse]:
    """
    Get blocked users by authenticated user

    **Scope**: ``profile:read``

    **Example requests**:

    - without parameters:

    .. sourcecode:: http

      GET /api/auth/blocked-users HTTP/1.1

    - with parameters:

    .. sourcecode:: http

      GET /api/auth/blocked-users?page=1
        HTTP/1.1

    **Example responses**:

    - with blocked users:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
          "blocked_users": [
            {
              "blocked": true,
              "created_at": "Sun, 01 Dec 2024 17:27:49 GMT",
              "followers": 0,
              "following": 0,
              "follows": "false",
              "is_followed_by": "false",
              "nb_workouts": 1,
              "picture": false,
              "role": "user",
              "suspended_at": null,
              "username": "Sam"
            }
          ],
          "pagination": {
            "has_next": false,
            "has_prev": false,
            "page": 1,
            "pages": 1,
            "total": 1
          },
          "status": "success"
        }

    - no blocked users:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
          "blocked_users": [],
          "pagination": {
            "has_next": false,
            "has_prev": false,
            "page": 1,
            "pages": 0,
            "total": 0
          },
          "status": "success"
        }

    :query integer page: page if using pagination (default: 1)

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``
    """
    params = request.args.copy()
    try:
        page = int(params.get("page", 1))
    except ValueError:
        page = 1

    paginated_relations = (
        User.query.join(BlockedUser, User.id == BlockedUser.user_id)
        .filter(BlockedUser.by_user_id == auth_user.id)
        .order_by(BlockedUser.created_at.desc())
        .paginate(page=page, per_page=BLOCKED_USERS_PER_PAGE, error_out=False)
    )
    return {
        "status": "success",
        "blocked_users": [
            user.serialize(current_user=auth_user)
            for user in paginated_relations.items
        ],
        "pagination": {
            "has_next": paginated_relations.has_next,
            "has_prev": paginated_relations.has_prev,
            "page": paginated_relations.page,
            "pages": paginated_relations.pages,
            "total": paginated_relations.total,
        },
    }


@auth_blueprint.route("/auth/account/suspension", methods=["GET"])
@require_auth(scopes=["profile:read"], allow_suspended_user=True)
def get_user_suspension(
    auth_user: User,
) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Get suspension if exists for authenticated user.

    Suspended user can access this endpoint.

    **Scope**: ``profile:read``

    **Example request**:

    .. sourcecode:: http

      GET /api/auth/account/suspension HTTP/1.1

    **Example responses**:

    - suspension exists:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
          "status": "success",
          "user_suspension": {
            "action_type": "user_suspension",
            "appeal": null,
            "comment": null,
            "created_at": "Wed, 04 Dec 2024 10:45:13 GMT",
            "id": "mmy3qPL3vcFuKJGfFBnCJV",
            "reason": "<SUSPENSION REASON>",
            "workout": null
          }
        }

    - no suspension:

    .. sourcecode:: http

      HTTP/1.1 404 NOT FOUND
      Content-Type: application/json

        {
          "status": "not found",
          "message": "user account is not suspended"
        }

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 404: ``user account is not suspended``
    """
    if auth_user.suspended_at is None or auth_user.suspension_action is None:
        return NotFoundErrorResponse("user account is not suspended")

    return {
        "status": "success",
        "user_suspension": auth_user.suspension_action.serialize(auth_user),
    }, 200


@auth_blueprint.route(
    "/auth/account/suspension/appeal",
    methods=["POST"],
)
@require_auth(scopes=["profile:write"], allow_suspended_user=True)
def appeal_user_suspension(
    auth_user: User,
) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Appeal suspension for authenticated user.

    Suspended user can access this endpoint.

    **Scope**: ``profile:write``

    **Example request**:

    .. sourcecode:: http

      POST /api/auth/account/suspension/appeal HTTP/1.1

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 201 CREATED
      Content-Type: application/json

        {
          "status": "success"
        }

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :<json string text: text explaining appeal

    :statuscode 201: appeal for suspension created
    :statuscode 400:
        - ``no text provided``
        - ``you can appeal only once``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 404: ``user account is not suspended``
    :statuscode 500: ``error, please try again or contact the administrator``
    """
    if auth_user.suspended_at is None or auth_user.suspension_action is None:
        return NotFoundErrorResponse("user account is not suspended")

    text = request.get_json().get("text")
    if not text:
        return InvalidPayloadErrorResponse("no text provided")

    try:
        appeal = ReportActionAppeal(
            action_id=auth_user.suspension_action.id,
            user_id=auth_user.id,
            text=text,
        )
        db.session.add(appeal)
        db.session.commit()
        return {"status": "success"}, 201

    except exc.IntegrityError:
        return InvalidPayloadErrorResponse("you can appeal only once")
    except (exc.OperationalError, ValueError) as e:
        return handle_error_and_return_response(e, db=db)


@auth_blueprint.route(
    "/auth/account/sanctions/<string:action_short_id>", methods=["GET"]
)
@require_auth(scopes=["profile:read"], allow_suspended_user=True)
def get_user_sanction(
    auth_user: User, action_short_id: str
) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Get sanction for authenticated user.

    Suspended user can access this endpoint.

    **Scope**: ``profile:read``

    **Example request**:

    .. sourcecode:: http

      GET /api/auth/account/sanctions/mmy3qPL3vcFuKJGfFBnCJV HTTP/1.1

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 SUCCESS
      Content-Type: application/json

        {
          "sanction": {
            "action_type": "user_suspension",
            "appeal": {
              "approved": null,
              "created_at": "Wed, 04 Dec 2024 10:49:00 GMT",
              "id": "7pDujhCVHyA4hv29JZQNgg",
              "reason": null,
              "text": "<APPEAL TEXT>",
              "updated_at": null
            },
            "comment": null,
            "created_at": "Wed, 04 Dec 2024 10:45:13 GMT",
            "id": "mmy3qPL3vcFuKJGfFBnCJV",
            "reason": "<SANCTION REASON>",
            "workout": null
          },
          "status": "success"
        }

    :param string action_short_id: suspension id

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 404: ``no sanction found``
    """
    sanction = ReportAction.query.filter_by(
        uuid=decode_short_id(action_short_id), user_id=auth_user.id
    ).first()

    if not sanction:
        return NotFoundErrorResponse("no sanction found")

    return {
        "status": "success",
        "sanction": sanction.serialize(current_user=auth_user, full=True),
    }, 200


@auth_blueprint.route(
    "/auth/account/sanctions/<string:action_short_id>/appeal",
    methods=["POST"],
)
@require_auth(scopes=["profile:write"])
def appeal_user_sanction(
    auth_user: User, action_short_id: str
) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Appeal a sanction

    **Scope**: ``profile:write``

    **Example request**:

    .. sourcecode:: http

      POST /api/auth/account/sanctions/6dxczvMrhkAR72shUz9Pwd/appeal HTTP/1.1

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 201 CREATED
      Content-Type: application/json

        {
          "status": "success"
        }

    :param string action_short_id: sanction id

    :<json string text: text explaining appeal

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 201: appeal created
    :statuscode 400:
        - ``no text provided``
        - ``you can appeal only once``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``
    :statuscode 404: ``no sanction found``
    :statuscode 500: ``error, please try again or contact the administrator``
    """
    sanction = ReportAction.query.filter_by(
        uuid=decode_short_id(action_short_id), user_id=auth_user.id
    ).first()

    if not sanction:
        return NotFoundErrorResponse("no sanction found")

    if sanction.report.object_type == "comment" and not sanction.comment_id:
        return InvalidPayloadErrorResponse("comment has been deleted")

    if sanction.report.object_type == "workout" and not sanction.workout_id:
        return InvalidPayloadErrorResponse("workout has been deleted")

    text = request.get_json().get("text")
    if not text:
        return InvalidPayloadErrorResponse("no text provided")

    try:
        appeal = ReportActionAppeal(
            action_id=sanction.id,
            user_id=auth_user.id,
            text=text,
        )
        db.session.add(appeal)
        db.session.commit()
        return {"status": "success"}, 201

    except exc.IntegrityError:
        return InvalidPayloadErrorResponse("you can appeal only once")
    except (exc.OperationalError, ValueError) as e:
        return handle_error_and_return_response(e, db=db)


@auth_blueprint.route("/auth/timezones", methods=["GET"])
def get_timezones() -> Tuple[Dict, int]:
    """
    Returns list of available time zones

    **Example request**:

    .. sourcecode:: http

      GET /api/auth/timezones HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "timezones": [
          "Africa/Abidjan",
          "Africa/Accra",
          "Africa/Algiers",
          "Africa/Bissau",
          "Africa/Cairo",
          "...",
          "Pacific/Tahiti",
          "Pacific/Tarawa",
          "Pacific/Tongatapu",
          "Pacific/Wake",
          "Pacific/Wallis",
        ],
        "status": "success"
      }

    :statuscode 200: ``success``
    """
    return {"timezones": TIMEZONES, "status": "success"}, 200



from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from statistics import mean
from typing import IO, TYPE_CHECKING, Dict, List, Optional, Tuple, Union

import gpxpy.gpx
from lxml import etree as ET

from fittrackee import db

from ...exceptions import WorkoutExceedingValueException, WorkoutFileException
from ...models import WORKOUT_VALUES_LIMIT, Workout, WorkoutSegment
from .base_workout_with_segment_service import (
    BaseWorkoutWithSegmentsCreationService,
)
from .workout_point import WorkoutPoint

if TYPE_CHECKING:
    from uuid import UUID

    from fittrackee.users.models import User


@dataclass
class GpxInfo:
    distance: float
    duration: Optional[float]
    max_speed: float
    moving_time: float
    stopped_time: float
    max_alt: Optional[float]
    min_alt: Optional[float]
    ascent: Optional[float] = None
    descent: Optional[float] = None


def remove_microseconds(delta: "timedelta") -> "timedelta":
    return delta - timedelta(microseconds=delta.microseconds)


class WorkoutGpxCreationService(BaseWorkoutWithSegmentsCreationService):
    def __init__(
        self,
        auth_user: "User",
        workout_file: IO[bytes],
        sport_id: int,
        stopped_speed_threshold: float,
        get_weather: bool = True,
    ):
        super().__init__(
            auth_user,
            workout_file,
            sport_id,
            stopped_speed_threshold,
            get_weather,
        )
        self.gpx: "gpxpy.gpx.GPX" = self.parse_file(workout_file)
        self.cadences: List[int] = []
        self.heart_rates: List[int] = []

    @staticmethod
    def _get_extensions(
        heart_rate: Optional[int], cadence: Optional[int]
    ) -> "ET.Element":
        track_point_extension = ET.Element("{gpxtpx}TrackPointExtension")
        if heart_rate is not None:
            heart_rate_element = ET.SubElement(
                track_point_extension, "{gpxtpx}hr"
            )
            heart_rate_element.text = str(heart_rate)
        if cadence is not None:
            cadence_element = ET.SubElement(
                track_point_extension, "{gpxtpx}cad"
            )
            cadence_element.text = str(cadence)
        return track_point_extension

    @classmethod
    def parse_file(cls, workout_file: IO[bytes]) -> "gpxpy.gpx.GPX":
        try:
            gpx = gpxpy.parse(workout_file)  # type: ignore
        except Exception as e:
            raise WorkoutFileException(
                "error", "error when parsing gpx file"
            ) from e
        if len(gpx.tracks) == 0:
            raise WorkoutFileException(
                "error", "no tracks in gpx file"
            ) from None
        return gpx

    @staticmethod
    def get_gpx_info(
        *,
        parsed_gpx: Union["gpxpy.gpx.GPXTrack", "gpxpy.gpx.GPXTrackSegment"],
        stopped_speed_threshold: float,
        use_raw_gpx_speed: bool,
    ) -> GpxInfo:
        moving_data = parsed_gpx.get_moving_data(
            stopped_speed_threshold=stopped_speed_threshold,
            raw=use_raw_gpx_speed,
        )
        if not moving_data:
            raise WorkoutFileException(
                "error", "gpx file is invalid"
            ) from None
        elevation = parsed_gpx.get_elevation_extremes()
        gpx_info = GpxInfo(
            duration=parsed_gpx.get_duration(),
            distance=(
                moving_data.moving_distance + moving_data.stopped_distance
            ),
            moving_time=moving_data.moving_time,
            stopped_time=moving_data.stopped_time,
            max_speed=moving_data.max_speed,
            max_alt=elevation.maximum,
            min_alt=elevation.minimum,
        )
        if elevation.maximum:
            hill = parsed_gpx.get_uphill_downhill()
            gpx_info.ascent = hill.uphill
            gpx_info.descent = hill.downhill
        return gpx_info

    @staticmethod
    def check_gpx_info(gpx_info: "GpxInfo") -> None:
        for key, value in WORKOUT_VALUES_LIMIT.items():
            gpx_info_value = getattr(gpx_info, key)
            if gpx_info_value and gpx_info_value > value:
                raise WorkoutExceedingValueException(
                    f"'{key}' exceeds max value ({value})"
                )

    def set_calculated_data(
        self,
        *,
        parsed_gpx: Union["gpxpy.gpx.GPXTrack", "gpxpy.gpx.GPXTrackSegment"],
        object_to_update: Union["Workout", "WorkoutSegment"],
        stopped_time_between_segments: timedelta,
        stopped_speed_threshold: float,
        use_raw_gpx_speed: bool,
        hr_cadence_stats: dict,
    ) -> Union["Workout", "WorkoutSegment"]:
        gpx_info = self.get_gpx_info(
            parsed_gpx=parsed_gpx,
            stopped_speed_threshold=stopped_speed_threshold,
            use_raw_gpx_speed=use_raw_gpx_speed,
        )
        self.check_gpx_info(gpx_info)

        if isinstance(object_to_update, WorkoutSegment):
            object_to_update.max_speed = (gpx_info.max_speed / 1000) * 3600

        object_to_update.ascent = gpx_info.ascent
        object_to_update.ave_speed = (
            (
                gpx_info.distance / gpx_info.moving_time
                if gpx_info.moving_time > 0
                else 0
            )
            / 1000
            * 3600
        )
        object_to_update.descent = gpx_info.descent
        object_to_update.distance = gpx_info.distance / 1000
        object_to_update.duration = remove_microseconds(
            timedelta(seconds=gpx_info.duration if gpx_info.duration else 0)
            + stopped_time_between_segments
        )
        object_to_update.max_alt = gpx_info.max_alt
        object_to_update.min_alt = gpx_info.min_alt
        object_to_update.moving = remove_microseconds(
            timedelta(seconds=gpx_info.moving_time)
        )
        object_to_update.pauses = remove_microseconds(
            timedelta(seconds=gpx_info.stopped_time)
            + stopped_time_between_segments
        )

        object_to_update.ave_cadence = hr_cadence_stats["ave_cadence"]
        object_to_update.ave_hr = hr_cadence_stats["ave_hr"]
        object_to_update.max_cadence = hr_cadence_stats["max_cadence"]
        object_to_update.max_hr = hr_cadence_stats["max_hr"]

        return object_to_update

    def get_workout_date(self) -> "datetime":
        if not self.start_point or not self.start_point.time:
            raise WorkoutFileException(
                "error", "<time> is missing in gpx file"
            )
        return self.start_point.time.astimezone(timezone.utc)

    def _process_segment_points(
        self,
        points: List["gpxpy.gpx.GPXTrackPoint"],
        stopped_time_between_segments: timedelta,
        previous_segment_last_point_time: Optional[datetime],
        is_last_segment: bool,
    ) -> Tuple[timedelta, Optional[datetime], Dict]:
        last_point_index = len(points) - 1
        cadences = []
        heart_rates = []
        for point_idx, point in enumerate(points):
            if point_idx == 0:
                # if a previous segment exists, calculate stopped time
                # between the two segments
                if previous_segment_last_point_time and point.time:
                    stopped_time_between_segments += (
                        point.time - previous_segment_last_point_time
                    )

            if point.extensions:
                for element in point.extensions[0]:
                    if element.tag.endswith("}hr") and element.text:
                        heart_rates.append(int(element.text))
                    if element.tag.endswith("}cad") and element.text:
                        cadences.append(int(element.text))

            # last segment point
            if point_idx == last_point_index:
                previous_segment_last_point_time = point.time

                # last gpx point
                if is_last_segment and point.time:
                    self.end_point = WorkoutPoint(
                        point.longitude,
                        point.latitude,
                        point.time,
                    )
            self.coordinates.append([point.longitude, point.latitude])

        hr_cadence_stats = {
            "ave_cadence": mean(cadences) if cadences else None,
            "ave_hr": mean(heart_rates) if heart_rates else None,
            "max_cadence": max(cadences) if cadences else None,
            "max_hr": max(heart_rates) if heart_rates else None,
        }
        self.cadences.extend(cadences)
        self.heart_rates.extend(heart_rates)
        return (
            stopped_time_between_segments,
            previous_segment_last_point_time,
            hr_cadence_stats,
        )

    def _process_segments(
        self,
        segments: List["gpxpy.gpx.GPXTrackSegment"],
        new_workout_id: int,
        new_workout_uuid: "UUID",
    ) -> Tuple[timedelta, float]:
        last_segment_index = len(segments) - 1
        max_speed = 0.0
        previous_segment_last_point_time: Optional[datetime] = None
        stopped_time_between_segments = timedelta(seconds=0)
        for segment_idx, segment in enumerate(segments):
            new_workout_segment = WorkoutSegment(
                segment_id=segment_idx,
                workout_id=new_workout_id,
                workout_uuid=new_workout_uuid,
            )
            db.session.add(new_workout_segment)

            is_last_segment = segment_idx == last_segment_index
            (
                stopped_time_between_segments,
                previous_segment_last_point_time,
                hr_cadence_stats,
            ) = self._process_segment_points(
                segment.points,
                stopped_time_between_segments,
                previous_segment_last_point_time,
                is_last_segment,
            )

            self.set_calculated_data(
                parsed_gpx=segment,
                object_to_update=new_workout_segment,
                stopped_time_between_segments=timedelta(seconds=0),
                stopped_speed_threshold=self.stopped_speed_threshold,
                use_raw_gpx_speed=self.auth_user.use_raw_gpx_speed,
                hr_cadence_stats=hr_cadence_stats,
            )

            if (
                new_workout_segment.max_speed
                and new_workout_segment.max_speed > max_speed
            ):
                max_speed = new_workout_segment.max_speed
        return stopped_time_between_segments, max_speed

    def _process_file(self) -> "Workout":
        if not self.gpx:
            raise WorkoutFileException(
                "error", "no gpx, please load gpx file before"
            ) from None

        track: "gpxpy.gpx.GPXTrack" = self.gpx.tracks[0]
        start_point = track.segments[0].points[0]
        if start_point.time:
            self.start_point = WorkoutPoint(
                start_point.longitude,
                start_point.latitude,
                start_point.time,
            )
        self.workout_name = track.name
        self.workout_description = track.description

        new_workout = Workout(
            user_id=self.auth_user.id,
            sport_id=self.sport_id,
            workout_date=self.get_workout_date(),
        )
        db.session.add(new_workout)
        db.session.flush()
        new_workout.source = self.gpx.creator

        stopped_time_between_segments, max_speed = self._process_segments(
            track.segments, new_workout.id, new_workout.uuid
        )

        hr_cadence_stats = {
            "ave_cadence": mean(self.cadences) if self.cadences else None,
            "ave_hr": mean(self.heart_rates) if self.heart_rates else None,
            "max_cadence": max(self.cadences) if self.cadences else None,
            "max_hr": max(self.heart_rates) if self.heart_rates else None,
        }
        self.set_calculated_data(
            parsed_gpx=track,
            object_to_update=new_workout,
            stopped_time_between_segments=stopped_time_between_segments,
            stopped_speed_threshold=self.stopped_speed_threshold,
            use_raw_gpx_speed=self.auth_user.use_raw_gpx_speed,
            hr_cadence_stats=hr_cadence_stats,
        )
        new_workout.max_speed = max_speed
        bounds = track.get_bounds()
        new_workout.bounds = (
            [
                bounds.min_latitude,
                bounds.min_longitude,
                bounds.max_latitude,
                bounds.max_longitude,
            ]
            if (
                bounds
                and bounds.min_latitude
                and bounds.min_longitude
                and bounds.max_latitude
                and bounds.max_longitude
            )
            else []
        )

        return new_workout



from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Optional

from fittrackee.responses import (
    DataNotFoundErrorResponse,
    ForbiddenErrorResponse,
    NotFoundErrorResponse,
)

from .exceptions import WorkoutForbiddenException
from .utils.workouts import get_workout

if TYPE_CHECKING:
    from fittrackee.users.models import User


def check_workout(only_owner: bool = True, as_data: bool = True) -> Callable:
    def decorator_check_workout(f: Callable) -> Callable:
        @wraps(f)
        def wrapper_check_workout(*args: Any, **kwargs: Any) -> Callable:
            auth_user: Optional["User"] = args[0]
            workout_short_id = kwargs["workout_short_id"]
            try:
                workout = get_workout(workout_short_id, auth_user)
            except WorkoutForbiddenException:
                return (
                    DataNotFoundErrorResponse("workouts")
                    if as_data
                    else NotFoundErrorResponse(
                        f"workout not found (id: {workout_short_id})"
                    )
                )

            if only_owner and (
                not auth_user or auth_user.id != workout.user.id
            ):
                return ForbiddenErrorResponse()

            return f(auth_user, workout, **kwargs)

        return wrapper_check_workout

    return decorator_check_workout



from typing import Dict, Optional, Tuple, Union
from urllib.parse import parse_qsl

from flask import Blueprint, Response, request
from urllib3.util import parse_url

from fittrackee import db
from fittrackee.responses import (
    HttpResponse,
    InvalidPayloadErrorResponse,
    NotFoundErrorResponse,
)
from fittrackee.users.models import User

from .client import create_oauth2_client
from .exceptions import InvalidOAuth2Scopes
from .models import OAuth2Client, OAuth2Token
from .server import authorization_server, require_auth

oauth2_blueprint = Blueprint("oauth2", __name__)

EXPECTED_METADATA_KEYS = [
    "client_name",
    "client_uri",
    "redirect_uris",
    "scope",
]
DEFAULT_PER_PAGE = 5


def is_errored(url: str) -> Optional[str]:
    query = dict(parse_qsl(parse_url(url).query))
    if query.get("error"):
        return query.get("error_description", "invalid payload")
    return None


@oauth2_blueprint.route("/oauth/apps", methods=["GET"])
@require_auth(allow_suspended_user=True)
def get_clients(auth_user: User) -> Dict:
    """
    Get OAuth2 clients (apps) for authenticated user with pagination
    (5 clients/page).

    This endpoint is only accessible by FitTrackee client (first-party
    application).

    Suspended user can access this endpoint.

    **Example request**:

    - without parameters:

    .. sourcecode:: http

      GET /api/oauth/apps HTTP/1.1
      Content-Type: application/json

    - with 'page' parameter:

    .. sourcecode:: http

      GET /api/oauth/apps?page=2 HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 SUCCESS
      Content-Type: application/json

      {
        "data": {
          "clients": [
            {
              "client_description": "",
              "client_id": "o22a27s2aBPUoxJbxV3UjDOx",
              "id": 1,
              "issued_at": "Thu, 14 July 2022 06:27:53 GMT",
              "name": "GPX Importer",
              "redirect_uris": [
                " https://example.com/callback"
              ],
              "scope": "profile:read workouts:write",
              "website": "https://example.com"
            }
          ]
        },
        "pagination": {
          "has_next": false,
          "has_prev": false,
          "page": 1,
          "pages": 1,
          "total": 1
        },
        "status": "success"
      }

    :query integer page: page for pagination (default: 1)

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    """
    params = request.args.copy()
    page = int(params.get("page", 1))
    per_page = DEFAULT_PER_PAGE
    clients_pagination = (
        OAuth2Client.query.filter_by(user_id=auth_user.id)
        .order_by(OAuth2Client.id.desc())
        .paginate(page=page, per_page=per_page, error_out=False)
    )
    clients = clients_pagination.items
    return {
        "status": "success",
        "data": {
            "clients": [
                client.serialize(with_secret=False) for client in clients
            ]
        },
        "pagination": {
            "has_next": clients_pagination.has_next,
            "has_prev": clients_pagination.has_prev,
            "page": clients_pagination.page,
            "pages": clients_pagination.pages,
            "total": clients_pagination.total,
        },
    }


@oauth2_blueprint.route("/oauth/apps", methods=["POST"])
@require_auth()
def create_client(auth_user: User) -> Union[HttpResponse, Tuple[Dict, int]]:
    """
    Create an OAuth2 client (app) for the authenticated user.

    This endpoint is only accessible by FitTrackee client (first-party
    application).

    **Example request**:

    .. sourcecode:: http

      POST /api/oauth/apps HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 SUCCESS
      Content-Type: application/json

      {
        "data": {
          "client": {
            "client_description": "",
            "client_id": "o22a27s2aBPUoxJbxV3UjDOx",
            "client_secret": "<CLIENT SECRET>",
            "id": 1,
            "issued_at": "Thu, 14 July 2022 06:27:53 GMT",
            "name": "GPX Importer",
            "redirect_uris": [
              "https://example.com/callback"
            ],
            "scope": "profile:read workouts:write",
            "website": "https://example.com"
          }
        },
        "status": "created"
      }

    :json string client_name: client name
    :json string client_uri: client URL
    :json array  redirect_uri: list of client redirect URLs (string)
    :json string scope: client scopes
    :json string client_description: client description (optional)

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 400: ``invalid payload``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``
    """
    client_metadata = request.get_json()
    if not client_metadata:
        return InvalidPayloadErrorResponse(
            message="OAuth2 client metadata missing"
        )

    missing_keys = [
        key
        for key in EXPECTED_METADATA_KEYS
        if key not in client_metadata.keys()
    ]
    if missing_keys:
        return InvalidPayloadErrorResponse(
            message=(
                "OAuth2 client metadata missing keys: "
                f"{', '.join(missing_keys)}"
            )
        )

    try:
        new_client = create_oauth2_client(client_metadata, auth_user)
    except InvalidOAuth2Scopes:
        return InvalidPayloadErrorResponse(
            message="OAuth2 client invalid scopes"
        )

    db.session.add(new_client)
    db.session.commit()
    return (
        {
            "status": "created",
            "data": {"client": new_client.serialize(with_secret=True)},
        },
        201,
    )


def get_client(
    auth_user: User,
    client_id: Optional[int],
    client_client_id: Optional[str],
) -> Union[Dict, HttpResponse]:
    key = "id" if client_id else "client_id"
    value = client_id if client_id else client_client_id
    client = OAuth2Client.query.filter_by(
        **{key: value, "user_id": auth_user.id}
    ).first()

    if not client:
        return NotFoundErrorResponse("OAuth2 client not found")

    return {
        "status": "success",
        "data": {"client": client.serialize(with_secret=False)},
    }


@oauth2_blueprint.route(
    "/oauth/apps/<string:client_client_id>", methods=["GET"]
)
@require_auth(allow_suspended_user=True)
def get_client_by_client_id(
    auth_user: User, client_client_id: str
) -> Union[Dict, HttpResponse]:
    """
    Get an OAuth2 client (app) by 'client_id'.

    This endpoint is only accessible by FitTrackee client (first-party
    application).

    Suspended user can access this endpoint.

    **Example request**:

    .. sourcecode:: http

      GET /api/oauth/apps/o22a27s2aBPUoxJbxV3UjDOx HTTP/1.1
      Content-Type: application/json

    **Example responses**:

    - success:

    .. sourcecode:: http

      HTTP/1.1 200 SUCCESS
      Content-Type: application/json

      {
        "data": {
          "client": {
            "client_description": "",
            "client_id": "o22a27s2aBPUoxJbxV3UjDOx",
            "id": 1,
            "issued_at": "Thu, 14 July 2022 06:27:53 GMT",
            "name": "GPX Importer",
            "redirect_uris": [
              "https://example.com/callback"
            ],
            "scope": "profile:read workouts:write",
            "website": "https://example.com"
          }
        },
        "status": "success"
      }

    - not found:

    .. sourcecode:: http

      HTTP/1.1 404 NOT FOUND
      Content-Type: application/json

      {
        "status": "not found",
        "message": "OAuth2 client not found"
      }

    :param string client_client_id: OAuth2 client client_id

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 404: ``OAuth2 client not found``
    """
    return get_client(
        auth_user, client_id=None, client_client_id=client_client_id
    )


@oauth2_blueprint.route("/oauth/apps/<int:client_id>/by_id", methods=["GET"])
@require_auth(allow_suspended_user=True)
def get_client_by_id(
    auth_user: User, client_id: int
) -> Union[Dict, HttpResponse]:
    """
    Get an OAuth2 client (app) by id (integer value).

    This endpoint is only accessible by FitTrackee client (first-party
    application).

    Suspended user can access this endpoint.

    **Example request**:

    .. sourcecode:: http

      GET /api/oauth/apps/1/by_id HTTP/1.1
      Content-Type: application/json

    **Example responses**:

    - success:

    .. sourcecode:: http

      HTTP/1.1 200 SUCCESS
      Content-Type: application/json

      {
        "data": {
          "client": {
            "client_description": "",
            "client_id": "o22a27s2aBPUoxJbxV3UjDOx",
            "id": 1,
            "issued_at": "Thu, 14 July 2022 06:27:53 GMT",
            "name": "GPX Importer",
            "redirect_uris": [
              "https://example.com/callback"
            ],
            "scope": "profile:read workouts:write",
            "website": "https://example.com"
          }
        },
        "status": "success"
      }

    - not found:

    .. sourcecode:: http

      HTTP/1.1 404 NOT FOUND
      Content-Type: application/json

      {
        "status": "not found",
        "message": "OAuth2 client not found"
      }

    :param integer client_id: OAuth2 client id

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 404: ``OAuth2 client not found``
    """
    return get_client(auth_user, client_id=client_id, client_client_id=None)


@oauth2_blueprint.route("/oauth/apps/<int:client_id>", methods=["DELETE"])
@require_auth(allow_suspended_user=True)
def delete_client(
    auth_user: User, client_id: int
) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Delete an OAuth2 client (app).

    This endpoint is only accessible by FitTrackee client (first-party
    application).

    Suspended user can access this endpoint.

    **Example request**:

    .. sourcecode:: http

      DELETE /api/oauth/apps/1 HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 204 NO CONTENT
      Content-Type: application/json

    :param integer client_id: OAuth2 client id

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 204: OAuth2 client deleted
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 404: ``OAuth2 client not found``
    """
    client = OAuth2Client.query.filter_by(
        id=client_id,
        user_id=auth_user.id,
    ).first()

    if not client:
        return NotFoundErrorResponse("OAuth2 client not found")

    db.session.delete(client)
    db.session.commit()
    return {"status": "no content"}, 204


@oauth2_blueprint.route("/oauth/apps/<int:client_id>/revoke", methods=["POST"])
@require_auth(allow_suspended_user=True)
def revoke_client_tokens(
    auth_user: User, client_id: int
) -> Union[Dict, HttpResponse]:
    """
    Revoke all tokens associated to an OAuth2 client (app).

    This endpoint is only accessible by FitTrackee client (first-party
    application).

    Suspended user can access this endpoint.

    **Example request**:

    .. sourcecode:: http

      POST /api/oauth/apps/1/revoke HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 SUCCESS
      Content-Type: application/json

      {
        "status": "success"
      }

    :param integer client_id: OAuth2 client id

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 404: ``OAuth2 client not found``
    """
    client = OAuth2Client.query.filter_by(id=client_id).first()

    if not client:
        return NotFoundErrorResponse("OAuth2 client not found")

    OAuth2Token.revoke_client_tokens(client.client_id)
    return {"status": "success"}


@oauth2_blueprint.route("/oauth/authorize", methods=["POST"])
@require_auth()
def authorize(auth_user: User) -> Union[HttpResponse, Dict]:
    """
    Authorize an OAuth2 client (app).
    If successful, it redirects to the client callback URL with the code to
    issue a token.

    This endpoint is only accessible by FitTrackee client (first-party
    application).

    **Example request**:

    .. sourcecode:: http

      POST /api/oauth/authorize HTTP/1.1
      Content-Type: multipart/form-data

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 SUCCESS
      Content-Type: application/json

      {
        "status": "success"
      }

    :form string  client_id: OAuth2 client 'client_id'
    :form string  response_type: client response type (only 'code' is supported
                                by FitTrackee)
    :form string  scopes: OAuth2 client scopes
    :form boolean confirm: confirmation (must be ``true``)
    :form string  state: unique value to prevent cross-site request forgery
                         (not mandatory but recommended)
    :form string  code_challenge: string generated from a code verifier
                         (for PKCE, not mandatory but recommended)
    :form string  code_challenge_method: method used to create challenge,
                         for instance "S256" (mandatory if `code_challenge`
                         provided)

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 400:
        - ``invalid payload``
        - errors returned by Authlib library
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``
    """
    data = request.form
    if (
        not data
        or "client_id" not in data
        or "response_type" not in data
        or data.get("response_type") != "code"
    ):
        return InvalidPayloadErrorResponse()

    confirm = data.get("confirm", "false")
    grant_user = auth_user if confirm.lower() == "true" else None
    response = authorization_server.create_authorization_response(
        grant_user=grant_user
    )
    error_message = is_errored(url=response.location)
    if error_message:
        return InvalidPayloadErrorResponse(error_message)
    return {"redirect_url": response.location}


@oauth2_blueprint.route("/oauth/token", methods=["POST"])
def issue_token() -> Response:
    """
    Issue or refresh token for a given OAuth2 client (app).

    **Example request**:

    .. sourcecode:: http

      POST /api/oauth/token HTTP/1.1
      Content-Type: multipart/form-data

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 SUCCESS
      Content-Type: application/json

      {
        "access_token": "rOEHv64THCG28WcewZHRnVLUsOdUvw8NVnHKCmL57e",
        "expires_in": 864000,
        "refresh_token": "NuV9cY8VQOnrQKHTZ5pQAq2Zw7mSH0MorNPJr14AmSwD6f6I",
        "scope": ["profile:read", "workouts:write"],
        "token_type": "Bearer",
        "expires_at": 1658660147.0667062
      }

    :form string  client_id: OAuth2 client 'client_id'
    :form string  client_secret: OAuth2 client secret
    :form string  grant_type: OAuth2 client grant type
                         (only 'authorization_code' (for token issue)
                         and 'refresh_token' (for token refresh)
                         are supported by FitTrackee)
    :form string  code: code generated after authorizing the client
                         (for token issue)
    :form string  code_verifier: code verifier
                         (for token issue with PKCE, not mandatory)
    :form string  refresh_token: refresh token (for token refresh)

    :statuscode 200: ``success``
    :statuscode 400: errors returned by Authlib library
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    """
    return authorization_server.create_token_response()


@oauth2_blueprint.route("/oauth/revoke", methods=["POST"])
def revoke_token() -> Response:
    """
    Revoke a token for a given OAuth2 client (app).

    **Example request**:

    .. sourcecode:: http

      POST /api/oauth/revoke HTTP/1.1
      Content-Type: multipart/form-data

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 SUCCESS
      Content-Type: application/json

      {}

    :form string client_id: OAuth2 client 'client_id'
    :form string client_secret: OAuth2 client secret
    :form string token: access token to revoke

    :statuscode 200: ``success``
    :statuscode 400: errors returned by Authlib library
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    """
    return authorization_server.create_endpoint_response("revocation")



import time
from typing import Any, Dict, Optional

from authlib.integrations.sqla_oauth2 import (
    OAuth2AuthorizationCodeMixin,
    OAuth2ClientMixin,
    OAuth2TokenMixin,
)
from sqlalchemy.engine.base import Connection
from sqlalchemy.event import listens_for
from sqlalchemy.orm.mapper import Mapper
from sqlalchemy.orm.session import Session
from sqlalchemy.sql import text

from fittrackee import BaseModel, db


class OAuth2Client(BaseModel, OAuth2ClientMixin):
    __tablename__ = "oauth2_client"

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(
        db.Integer, db.ForeignKey("users.id", ondelete="CASCADE"), index=True
    )
    user = db.relationship("User")

    def serialize(self, with_secret: bool = False) -> Dict:
        client = {
            "client_id": self.client_id,
            "client_description": self.client_description,
            "id": self.id,
            "issued_at": time.strftime(
                "%a, %d %B %Y %H:%M:%S GMT",
                time.gmtime(self.client_id_issued_at),
            ),
            "name": self.client_name,
            "redirect_uris": self.redirect_uris,
            "scope": self.scope,
            "website": self.client_uri,
        }
        if with_secret:
            client["client_secret"] = self.client_secret
        return client

    @property
    def client_description(self) -> Optional[str]:
        return self.client_metadata.get("client_description")


@listens_for(OAuth2Client, "after_delete")
def on_old_oauth2_delete(
    mapper: Mapper, connection: Connection, old_oauth2_client: OAuth2Client
) -> None:
    @listens_for(db.Session, "after_flush", once=True)
    def receive_after_flush(session: Session, context: Any) -> None:
        session.query(OAuth2AuthorizationCode).filter(
            OAuth2AuthorizationCode.client_id == old_oauth2_client.client_id
        ).delete(synchronize_session=False)
        session.query(OAuth2Token).filter(
            OAuth2Token.client_id == old_oauth2_client.client_id
        ).delete(synchronize_session=False)


class OAuth2AuthorizationCode(BaseModel, OAuth2AuthorizationCodeMixin):
    __tablename__ = "oauth2_code"
    __table_args__ = (
        db.Index(
            "ix_oauth2_code_client_id",
            "client_id",
        ),
    )

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(
        db.Integer, db.ForeignKey("users.id", ondelete="CASCADE"), index=True
    )
    user = db.relationship("User")


class OAuth2Token(BaseModel, OAuth2TokenMixin):
    __tablename__ = "oauth2_token"

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(
        db.Integer, db.ForeignKey("users.id", ondelete="CASCADE"), index=True
    )
    user = db.relationship("User")

    def is_refresh_token_active(self) -> bool:
        if self.is_revoked():
            return False
        expires_at = self.issued_at + self.expires_in * 2
        return expires_at >= time.time()

    @classmethod
    def revoke_client_tokens(cls, client_id: str) -> None:
        sql = """
            UPDATE oauth2_token
            SET access_token_revoked_at = :revoked_at
            WHERE client_id = :client_id;
        """
        db.session.execute(
            text(sql), {"client_id": client_id, "revoked_at": int(time.time())}
        )
        db.session.commit()



from datetime import timezone
from typing import List, Optional

import gpxpy.gpx

from ..constants import CADENCE_SPORTS
from ..exceptions import WorkoutGPXException


def open_gpx_file(gpx_file: str) -> Optional[gpxpy.gpx.GPX]:
    gpx_file = open(gpx_file, "r")  # type: ignore
    gpx = gpxpy.parse(gpx_file)
    if len(gpx.tracks) == 0:
        return None
    return gpx


def get_gpx_segments(
    track_segments: List, segment_id: Optional[int] = None
) -> List:
    """
    Return list of segments, filtered on segment id if provided
    """
    if segment_id is not None:
        segment_index = segment_id - 1
        if segment_index > (len(track_segments) - 1):
            raise WorkoutGPXException(
                "not found", f"No segment with id '{segment_id}'", None
            )
        if segment_index < 0:
            raise WorkoutGPXException("error", "Incorrect segment id", None)
        segments = [track_segments[segment_index]]
    else:
        segments = track_segments

    return segments


def get_chart_data(
    gpx_file: str,
    sport_label: str,
    *,
    can_see_heart_rate: bool,
    segment_id: Optional[int] = None,
) -> Optional[List]:
    """
    Return data needed to generate chart with:
    - speed
    - elevation (if available)
    - heart rate (if available)
    - cadence (if available)
    """
    gpx = open_gpx_file(gpx_file)
    if gpx is None:
        return None

    chart_data = []
    first_point = None
    previous_point = None
    previous_distance = 0
    return_cadence = sport_label in CADENCE_SPORTS

    track_segments = gpx.tracks[0].segments
    segments = get_gpx_segments(track_segments, segment_id)

    for segment_idx, segment in enumerate(segments):
        for point_idx, point in enumerate(segment.points):
            if segment_idx == 0 and point_idx == 0:
                first_point = point
            distance = (
                point.distance_3d(previous_point)
                if (
                    point.elevation
                    and previous_point
                    and previous_point.elevation
                )
                else point.distance_2d(previous_point)
            )
            distance = 0 if distance is None else distance
            distance += previous_distance
            speed = (
                round((segment.get_speed(point_idx) / 1000) * 3600, 2)
                if segment.get_speed(point_idx) is not None
                else 0
            )
            data = {
                "distance": (
                    round(distance / 1000, 2) if distance is not None else 0
                ),
                "duration": int(point.time_difference(first_point)),
                "latitude": point.latitude,
                "longitude": point.longitude,
                "speed": speed,
                # workaround
                # https://github.com/tkrajina/gpxpy/issues/209
                "time": point.time.replace(
                    tzinfo=timezone(point.time.utcoffset())
                ),
            }
            if point.elevation:
                data["elevation"] = round(point.elevation, 1)
            if point.extensions:
                extensions = []
                for extension in point.extensions:
                    if "TrackPointExtension" in extension.tag:
                        extensions.extend(extension)
                    else:
                        extensions.append(extension)
                for element in extensions:
                    if (
                        can_see_heart_rate
                        and element.tag.endswith("hr")
                        and element.text
                    ):
                        data["hr"] = int(element.text)
                    if (
                        return_cadence
                        and (
                            element.tag.endswith("cad")
                            or element.tag.endswith("cadence")
                        )
                        and element.text
                    ):
                        data["cadence"] = int(element.text)
            chart_data.append(data)
            previous_point = point
            previous_distance = distance

    return chart_data


def extract_segment_from_gpx_file(
    content: str, segment_id: int
) -> Optional[str]:
    """
    Returns segments in xml format from a gpx file content
    """
    gpx_content = gpxpy.parse(content)
    if len(gpx_content.tracks) == 0:
        return None

    track_segment = get_gpx_segments(
        gpx_content.tracks[0].segments, segment_id
    )

    gpx = gpxpy.gpx.GPX()
    gpx_track = gpxpy.gpx.GPXTrack()
    gpx.tracks.append(gpx_track)
    gpx_segment = gpxpy.gpx.GPXTrackSegment()
    gpx_track.segments.append(gpx_segment)

    for point in track_segment[0].points:
        gpx_segment.points.append(
            gpxpy.gpx.GPXTrackPoint(
                point.latitude, point.longitude, elevation=point.elevation
            )
        )

    return gpx.to_xml()



"""add index on workout_equipments

Revision ID: 0ea28f1e6c60
Revises: 171dcd7e5f2b
Create Date: 2024-04-21 16:50:26.813151

"""

from alembic import op


# revision identifiers, used by Alembic.
revision = '0ea28f1e6c60'
down_revision = '171dcd7e5f2b'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('workout_equipments', schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f('ix_workout_equipments_equipment_id'),
            ['equipment_id'],
            unique=False,
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('workout_equipments', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_workout_equipments_equipment_id'))

    # ### end Alembic commands ###



WORKOUT_ALLOWED_EXTENSIONS = {"gpx", "fit", "kml", "kmz", "tcx"}
WORKOUT_DATE_FORMAT = "%Y-%m-%d %H:%M"
CADENCE_SPORTS = [
    "Cycling (Sport)",
    "Cycling (Trekking)",
    "Cycling (Transport)",
    "Cycling (Virtual)",
    "Halfbike",
    "Mountain Biking",
    "Mountain Biking (Electric)",
    "Hiking",
    "Mountaineering",
    "Snowshoes",
    "Running",
    "Trail",
    "Walking",
    "Open Water Swimming",
]



"""Update Nullable after sqlalchemy upgrade

Revision ID: 070bd31919c3
Revises: d3c5338bedff
Create Date: 2025-01-26 08:52:44.942137

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '070bd31919c3'
down_revision = 'd3c5338bedff'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('blocked_users', schema=None) as batch_op:
        batch_op.alter_column('user_id',
               existing_type=sa.INTEGER(),
               nullable=False)
        batch_op.alter_column('by_user_id',
               existing_type=sa.INTEGER(),
               nullable=False)

    with op.batch_alter_table('comments', schema=None) as batch_op:
        batch_op.alter_column('created_at',
               existing_type=postgresql.TIMESTAMP(),
               nullable=False)

    with op.batch_alter_table('equipments', schema=None) as batch_op:
        batch_op.alter_column('equipment_type_id',
               existing_type=sa.INTEGER(),
               nullable=False)
        batch_op.alter_column('creation_date',
               existing_type=postgresql.TIMESTAMP(),
               nullable=False)
        batch_op.alter_column('total_distance',
               existing_type=sa.NUMERIC(precision=10, scale=3),
               nullable=False,
               existing_server_default=sa.text('0.0'))

    with op.batch_alter_table('notifications', schema=None) as batch_op:
        batch_op.alter_column('from_user_id',
               existing_type=sa.INTEGER(),
               nullable=False)
        batch_op.alter_column('to_user_id',
               existing_type=sa.INTEGER(),
               nullable=False)

    with op.batch_alter_table('records', schema=None) as batch_op:
        batch_op.alter_column('record_type',
               existing_type=postgresql.ENUM('AS', 'FD', 'LD', 'MS', 'HA', name='record_types'),
               nullable=False)

    with op.batch_alter_table('report_actions', schema=None) as batch_op:
        batch_op.alter_column('created_at',
               existing_type=postgresql.TIMESTAMP(),
               nullable=False)

    with op.batch_alter_table('report_comments', schema=None) as batch_op:
        batch_op.alter_column('created_at',
               existing_type=postgresql.TIMESTAMP(),
               nullable=False)

    with op.batch_alter_table('reports', schema=None) as batch_op:
        batch_op.alter_column('created_at',
               existing_type=postgresql.TIMESTAMP(),
               nullable=False)

    with op.batch_alter_table('users_data_export', schema=None) as batch_op:
        batch_op.alter_column('user_id',
               existing_type=sa.INTEGER(),
               nullable=False)

    with op.batch_alter_table('workouts', schema=None) as batch_op:
        batch_op.alter_column('creation_date',
               existing_type=postgresql.TIMESTAMP(),
               nullable=False)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('workouts', schema=None) as batch_op:
        batch_op.alter_column('creation_date',
               existing_type=postgresql.TIMESTAMP(),
               nullable=True)

    with op.batch_alter_table('users_data_export', schema=None) as batch_op:
        batch_op.alter_column('user_id',
               existing_type=sa.INTEGER(),
               nullable=True)

    with op.batch_alter_table('reports', schema=None) as batch_op:
        batch_op.alter_column('created_at',
               existing_type=postgresql.TIMESTAMP(),
               nullable=True)

    with op.batch_alter_table('report_comments', schema=None) as batch_op:
        batch_op.alter_column('created_at',
               existing_type=postgresql.TIMESTAMP(),
               nullable=True)

    with op.batch_alter_table('report_actions', schema=None) as batch_op:
        batch_op.alter_column('created_at',
               existing_type=postgresql.TIMESTAMP(),
               nullable=True)

    with op.batch_alter_table('records', schema=None) as batch_op:
        batch_op.alter_column('record_type',
               existing_type=postgresql.ENUM('AS', 'FD', 'LD', 'MS', 'HA', name='record_types'),
               nullable=True)

    with op.batch_alter_table('notifications', schema=None) as batch_op:
        batch_op.alter_column('to_user_id',
               existing_type=sa.INTEGER(),
               nullable=True)
        batch_op.alter_column('from_user_id',
               existing_type=sa.INTEGER(),
               nullable=True)

    with op.batch_alter_table('equipments', schema=None) as batch_op:
        batch_op.alter_column('total_distance',
               existing_type=sa.NUMERIC(precision=10, scale=3),
               nullable=True,
               existing_server_default=sa.text('0.0'))
        batch_op.alter_column('creation_date',
               existing_type=postgresql.TIMESTAMP(),
               nullable=True)
        batch_op.alter_column('equipment_type_id',
               existing_type=sa.INTEGER(),
               nullable=True)

    with op.batch_alter_table('comments', schema=None) as batch_op:
        batch_op.alter_column('created_at',
               existing_type=postgresql.TIMESTAMP(),
               nullable=True)

    with op.batch_alter_table('blocked_users', schema=None) as batch_op:
        batch_op.alter_column('by_user_id',
               existing_type=sa.INTEGER(),
               nullable=True)
        batch_op.alter_column('user_id',
               existing_type=sa.INTEGER(),
               nullable=True)

    # ### end Alembic commands ###



"""add missing indexes on Workout table

Revision ID: e30007d681cb
Revises: ed409fd9db9d
Create Date: 2022-02-05 11:52:23.847975

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'e30007d681cb'
down_revision = 'ed409fd9db9d'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index(op.f('ix_workouts_map_id'), 'workouts', ['map_id'], unique=False)
    op.create_index(op.f('ix_workouts_sport_id'), 'workouts', ['sport_id'], unique=False)
    op.create_index(op.f('ix_workouts_user_id'), 'workouts', ['user_id'], unique=False)
    op.create_index(op.f('ix_workouts_workout_date'), 'workouts', ['workout_date'], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f('ix_workouts_workout_date'), table_name='workouts')
    op.drop_index(op.f('ix_workouts_user_id'), table_name='workouts')
    op.drop_index(op.f('ix_workouts_sport_id'), table_name='workouts')
    op.drop_index(op.f('ix_workouts_map_id'), table_name='workouts')
    # ### end Alembic commands ###



from .workout_creation_service import WorkoutCreationService
from .workout_from_file import (
    WorkoutFitCreationService,
    WorkoutGpxCreationService,
    WorkoutKmlCreationService,
    WorkoutKmzCreationService,
    WorkoutTcxCreationService,
)
from .workout_update_service import WorkoutUpdateService
from .workouts_from_file_creation_service import (
    WorkoutsFromArchiveCreationAsyncService,
    WorkoutsFromFileCreationService,
)

__all__ = [
    "WorkoutCreationService",
    "WorkoutFitCreationService",
    "WorkoutGpxCreationService",
    "WorkoutKmlCreationService",
    "WorkoutKmzCreationService",
    "WorkoutTcxCreationService",
    "WorkoutUpdateService",
    "WorkoutsFromArchiveCreationAsyncService",
    "WorkoutsFromFileCreationService",
]



import zipfile
from typing import IO, TYPE_CHECKING

from ...exceptions import WorkoutFileException
from .workout_kml_creation_service import WorkoutKmlCreationService

if TYPE_CHECKING:
    import gpxpy.gpx


class WorkoutKmzCreationService(WorkoutKmlCreationService):
    @classmethod
    def parse_file(cls, workout_file: IO[bytes]) -> "gpxpy.gpx.GPX":
        """
        Only kmz files with Tracks are supported.
        Note: for now kmz with photos are not supported.

        Tested with files generated with OpenTracks.
        """
        with zipfile.ZipFile(workout_file, "r") as kmz_ref:
            try:
                kml_content = kmz_ref.open("doc.kml")
            except KeyError as e:
                raise WorkoutFileException(
                    "error", "error when parsing kmz file"
                ) from e
        return super().parse_file(kml_content)



from typing import TYPE_CHECKING

from ..exceptions import WorkoutExceedingValueException
from ..models import PSQL_INTEGER_LIMIT, WORKOUT_VALUES_LIMIT

if TYPE_CHECKING:
    from fittrackee.workouts.models import Workout


class CheckWorkoutMixin:
    @staticmethod
    def _check_workout(workout: "Workout") -> None:
        if workout.distance and workout.distance > 999.9:
            raise WorkoutExceedingValueException(
                "'distance' exceeds max value (999.9)"
            )
        if workout.duration.total_seconds() > PSQL_INTEGER_LIMIT:
            raise WorkoutExceedingValueException(
                f"'duration' exceeds max value ({PSQL_INTEGER_LIMIT})"
            )

        for key in ["ascent", "descent", "max_speed"]:
            workout_value = getattr(workout, key)
            max_value = WORKOUT_VALUES_LIMIT[key]
            if workout_value and workout_value > max_value:
                raise WorkoutExceedingValueException(
                    f"'{key}' exceeds max value ({max_value})"
                )



import logging
import sys

import click

from fittrackee.cli.app import app
from fittrackee.workouts.tasks import (
    process_workouts_archive_upload,
    process_workouts_archives_uploads,
)

handler = logging.StreamHandler()
logger = logging.getLogger("fittrackee_workouts_cli")
logger.setLevel(logging.INFO)
logger.addHandler(handler)


@click.group(name="workouts")
def workouts_cli() -> None:
    """Manage workouts."""
    pass


@workouts_cli.command("archive_uploads")
@click.option(
    "--max",
    "max_archives",
    type=int,
    required=True,
    help="Maximum number of workouts archive to process.",
)
@click.option(
    "--verbose",
    "-v",
    "verbose",
    is_flag=True,
    default=False,
    help="Enable verbose output log (default: disabled).",
)
def workouts_archive_uploads(max_archives: int, verbose: bool) -> None:
    """
    Process workouts archive uploads if queued tasks exist (progress = 0 and
    not aborted/errored).
    To use in case redis is not set.
    """
    with app.app_context():
        if verbose:
            upload_logger = logging.getLogger("fittrackee_workouts_upload")
            upload_logger.setLevel(logging.DEBUG)
            upload_logger.addHandler(handler)
        count = process_workouts_archives_uploads(max_archives, logger)
        logger.info(f"\nWorkouts archives processed: {count}.")


@workouts_cli.command("archive_upload")
@click.option(
    "--id",
    "task_short_id",
    type=str,
    required=True,
    help="Id of task to process",
)
@click.option(
    "--verbose",
    "-v",
    "verbose",
    is_flag=True,
    default=False,
    help="Enable verbose output log (default: disabled).",
)
def process_queued_archive_upload(task_short_id: str, verbose: bool) -> None:
    """
    Process given queued workouts archive upload.

    To use in case redis is not set.
    """
    with app.app_context():
        if verbose:
            upload_logger = logging.getLogger("fittrackee_workouts_upload")
            upload_logger.setLevel(logging.DEBUG)
            upload_logger.addHandler(handler)
        try:
            process_workouts_archive_upload(task_short_id, logger)
        except Exception as e:
            logger.error(str(e))
            sys.exit(1)
        logger.info("\nDone.")



"""add stats_workouts_limit to AppConfig

Revision ID: 032ba0846eea
Revises: 0ea28f1e6c60
Create Date: 2024-06-19 10:27:12.825484

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '032ba0846eea'
down_revision = '0ea28f1e6c60'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('app_config', schema=None) as batch_op:
        batch_op.add_column(
            sa.Column('stats_workouts_limit', sa.Integer(), nullable=True)
        )

    op.execute("UPDATE app_config SET stats_workouts_limit = 10000")
    op.alter_column('app_config', 'stats_workouts_limit', nullable=False)


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('app_config', schema=None) as batch_op:
        batch_op.drop_column('stats_workouts_limit')

    # ### end Alembic commands ###



from .base_workout_with_segment_service import (
    BaseWorkoutWithSegmentsCreationService,
)
from .workout_fit_creation_service import WorkoutFitCreationService
from .workout_gpx_creation_service import GpxInfo, WorkoutGpxCreationService
from .workout_kml_creation_service import WorkoutKmlCreationService
from .workout_kmz_creation_service import WorkoutKmzCreationService
from .workout_point import WorkoutPoint
from .workout_tcx_creation_service import WorkoutTcxCreationService

__all__ = [
    "BaseWorkoutWithSegmentsCreationService",
    "GpxInfo",
    "WorkoutFitCreationService",
    "WorkoutGpxCreationService",
    "WorkoutKmlCreationService",
    "WorkoutKmzCreationService",
    "WorkoutPoint",
    "WorkoutTcxCreationService",
]



"""Add user preference for gpx speed calculation

Revision ID: eff1c16c43eb
Revises: db58d195c5bf
Create Date: 2023-05-14 22:12:56.244291

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'eff1c16c43eb'
down_revision = 'db58d195c5bf'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('users', schema=None) as batch_op:
        batch_op.add_column(sa.Column('use_raw_gpx_speed', sa.Boolean(), nullable=True))
    op.execute("UPDATE users SET use_raw_gpx_speed = false")
    op.alter_column('users', 'use_raw_gpx_speed', nullable=False)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('users', schema=None) as batch_op:
        batch_op.drop_column('use_raw_gpx_speed')

    # ### end Alembic commands ###
